from vai_lab._import_helper import import_module
from vai_lab.Data.xml_handler import XML_handler
from vai_lab.GUI.GUI_core import GUI
from vai_lab.Data.Data_core import Data
from vai_lab._plugin_helpers import PluginSpecs
from vai_lab._types import PluginSpecsInterface, ModuleInterface
import time
from sys import exit

from os.path import join
from typing import Dict, Tuple, Union, List
class Core:
    def __init__(self) -> None:
        self._xml_handler = XML_handler()
        self._avail_plugins: PluginSpecsInterface = PluginSpecs()
        self.data = Data()
        self.loop_level: int = 0
        self.setup_complete: bool = False
        
        self.status_logger:Dict = {}
        self._debug = False

    def _launch(self):
        gui_app = GUI()
        gui_app._debug = self._debug
        gui_app.set_avail_plugins(self._avail_plugins)
        gui_app.set_gui_as_startpage()
        gui_output = gui_app.launch()
        if not gui_app.closed:
            try:
                self.load_config_file(gui_output["xml_filename"])
            except:
                raise Exception("No XML File Selected. Cannot Run Pipeline")
            self._load_data()

    def load_config_file(self, filename: Union[str,List,Tuple]):
        if type(filename) == list or type(filename) == tuple:
            filedir:str = join(*filename)
        else:
            filedir = str(filename)
        self._xml_handler.load_XML(filedir)
        self.setup_complete = True

    def _load_data(self) -> None:
        init_data_fn = self._xml_handler.data_to_load
        self.data.import_data_from_config(init_data_fn)

    def _execute_module(self, specs):
        """Executes named module with given options
        Imports, instantiates, sets params, then launches module

        :param specs: dict of module to be executed
        """
        mod: ModuleInterface = import_module(globals(), specs["module_type"]).__call__()
        mod.set_avail_plugins(self._avail_plugins)
        mod.set_data_in(self.data)
        mod.set_options(specs)
        print("\t"*self.loop_level
                + specs["module_type"]
                + " module: \"{}\"".format(specs["name"])
                + "processing..."
              )
        mod.launch()
        self.data = mod.get_result()

    def _execute_loop(self, specs):
        try:
            print("\t"*self.loop_level
                        + specs["type"]
                        + " loop: "
                        + "\"{}\"".format(specs["name"])
                        + " starting...")
            self.loop_level += 1
            getattr(self, "_execute_{}_loop".format(specs["type"]))(specs)
            self.loop_level -= 1
        except KeyError:
            print("\nError: Invalid Loop Type.")
            print("Loop \"{0}\" with type \"{1}\" not recognised".format(
                specs.key, specs["type"]))

    def _execute_entry_point(self, specs):
        """Placeholder: Will parse the initialiser module when ready"""
        pass

    def _execute_exit_point(self, specs):
        """Placeholder: Will parse the Output module when ready"""
        pass

    def _parse_condition(self, condition):
        try:
            condition = int(condition)
            
            if isinstance(condition, int):
                return range(0, condition)
        except:
            print("Condition \"{0}\" cannot be parsed".format(condition))
            print(
                "Other formats in in development. Only ranged for loops are working currently")

    def _execute_for_loop(self, specs):
        condition = self._parse_condition(specs["condition"])
        for c in condition:
            self._execute(specs)

    def _runTracker(self):
        self.gui_app = GUI()
        self.gui_app.set_avail_plugins(self._avail_plugins)
        self.gui_app.set_gui(self.status_logger)
        self.gui_app._append_to_output("xml_filename", self._xml_handler.filename)
        return self.gui_app.launch()
    
    def _init_status(self, modules):
        for key in [key for key, val in modules.items() if type(val) == dict]:
            self.status_logger[modules[key]['name']] = {}

    def _add_status(self, module, key, value):
        self.status_logger[module['name']][key] = value

    def _progress_start(self, module):
        self._add_status(module, 'start', self._t)

    def _progress_finish(self, module):
        self._add_status(module, 'finish', self._t)


    @property
    def _t(self):
        return time.strftime('%H:%M:%S', time.localtime())

    def _execute(self, specs):
        """Run elements within a given dictionary.
        Only iterates over dict values that are dicts themselves.
        Non-dict elements cannot contain modules or loops

        :param specs: dict of elements to be executed
        """
        for key in [key for key, val in specs.items() if type(val) == dict]:
            
            self._progress_start(specs[key])
            getattr(self, "_execute_{}".format(specs[key]["class"]))(specs[key])
            self._progress_finish(specs[key])

            if specs[key]["class"] == 'module':
                term = self._runTracker()
                if not term['close']:
                    self.load_config_file(self._xml_handler.filename)
                else:
                    print('Pipeline terminated')
                    exit()

    def run(self):
        if not self.setup_complete:
            print("No pipeline specified. Running GUI.")
            print("To load existing config, run core.load_config_file(<path_to_file>)")
            self._launch()
        print("Running pipeline...")
        self._load_data()
        
        self._init_status(self._xml_handler.loaded_modules)
        self._execute(self._xml_handler.loaded_modules)
        print("Pipeline Complete")



import tkinter as tk
import os
from PIL import ImageTk
from tkinter import ttk
import numpy as np
from vai_lab._import_helper import get_lib_parent_dir
# from ttkwidgets import CheckboxTreeview
# from sys import platform

class dataLoader():
    """ Creates a window 
    """
    def __init__(self,controller,data):
        
        self.controller = controller
        self.newWindow = tk.Toplevel(self.controller)
        
        # Window options
        self.newWindow.title('Data Importer')
        self.newWindow.tk.call('wm','iconphoto', self.newWindow._w, ImageTk.PhotoImage(
            file = os.path.join(os.path.join(
                get_lib_parent_dir(), 
                'utils', 
                'resources', 
                'Assets', 
                'VAILabsIcon.ico'))))
        
        self.newWindow.geometry("700x400")
        self.newWindow.grid_rowconfigure(2, weight=1)
        self.newWindow.grid_columnconfigure(tuple(range(2)), weight=1)
        
        frame1 = tk.Frame(self.newWindow)
        # frame2 = tk.Frame(self.newWindow)
        tree_frame1 = tk.Frame(self.newWindow, height=300,width=300)
        tree_frame2 = tk.Frame(self.newWindow, height=300,width=300)
        frame3 = tk.Frame(self.newWindow)
        frame4 = tk.Frame(self.newWindow)
        
        sizegrip=ttk.Sizegrip(frame4)     # put a sizegrip widget on the southeast corner of the window
        sizegrip.pack(side = tk.RIGHT, anchor = tk.SE)

        tk.Label(frame1,
              text ="Select variables to import", 
              justify=tk.LEFT).pack()
        # tk.Label(frame2,
        #       text ="Variables in "+filename, 
        #       justify=tk.LEFT).pack()
        
        #Treeview 1
        style = ttk.Style()
        style.configure(
            "Treeview",
            background = 'white', 
            foreground = 'black', 
            rowheight = 25, 
            fieldbackground = 'white', 
            # font = self.controller.pages_font)
            )
        # style.configure("Treeview.Heading", 
        #                 # font = self.controller.pages_font)
        #                 )
        style.map('Treeview', 
                  background = [('selected', 'grey')])
        style.layout('cb.Treeview.Row',
                     [('Treeitem.row', {'sticky': 'nswe'}),
                      ('Treeitem.image', {'side': 'right', 'sticky': 'e'})])
        
        self.tree = []
        columns = ['Name', 'Size', 'Class']
        self.add_treeview(tree_frame1, data, columns)
        self.add_treeview(tree_frame2, None, [])
        tk.Button(
            frame3, text = 'Finish', fg = 'black', 
            height = 2, width = 10, #font = self.controller.pages_font, 
            command = lambda: self.check_quit()).pack(side = tk.RIGHT, anchor = tk.W)

        tree_frame1.grid(column=0, row=2, sticky="nsew")
        tree_frame2.grid(column=1, row=2, sticky="nsew")
        frame1.grid(column=0, row=0, columnspan=2, sticky="nw")
        # frame2.grid(column=0, row=1, columnspan=2, sticky="nw")
        frame3.grid(column=0, row=3, columnspan=2, sticky="n")
        frame4.grid(column=0, row=4, columnspan=2, sticky="se")

    def add_treeview(self,frame,data,columns):
        """ Add a treeview"""

        tree_scrollx = tk.Scrollbar(frame, orient = 'horizontal')
        tree_scrollx.pack(side = tk.BOTTOM, fill = tk.BOTH)
        tree_scrolly = tk.Scrollbar(frame)
        tree_scrolly.pack(side = tk.RIGHT, fill = tk.BOTH)
        if data is not None:
            self.tree.append(ttk.Treeview(
                frame, 
                yscrollcommand = tree_scrolly.set, 
                xscrollcommand = tree_scrollx.set,
                show=("headings", "tree")))
        else:
            self.tree.append(ttk.Treeview(
                frame, 
                yscrollcommand = tree_scrolly.set, 
                xscrollcommand = tree_scrollx.set))

        tree_scrollx.config(command = self.tree[-1].xview)
        tree_scrolly.config(command = self.tree[-1].yview)
        # self.tree[-1].pack_propagate(0)
        
        self.tree[-1].pack(expand = True, side = tk.LEFT, fill = tk.BOTH)
        self.fill_treeview(self.tree[-1], data, columns)

    def fill_treeview(self,tree,data,columns):
        if data is not None:
            tree['columns'] = columns

            # Format columns
            tree.column("#0", width = 0, stretch=tk.NO)
            for cl in tree['columns']:
                tree.column(
                    cl, width = int(
                        self.controller.pages_font.measure(str(cl)))+30, 
                    minwidth = 50, anchor = tk.CENTER, stretch=tk.NO)

            # Headings
            tree.heading("#0")
            for cl in tree['columns']:
                tree.heading(cl, text = cl, anchor = tk.CENTER)
            
            variables = [key for key in data.keys() if (key[:1] != '__') and (key[-2:] != '__')]
            for n, var in enumerate(variables):
                if n%2 == 0:
                    tree.insert(parent = '', index = 'end', iid = n, text = '', 
                                     values = (var, data[var].values.shape, data[var].values.dtype), tags = ('even',))
                else:
                    tree.insert(parent = '', index = 'end', iid = n, text = '', 
                                     values = (var, data[var].shape, data[var].values.dtype), tags = ('odd',))

            # Define double-click on row action
            if len(self.tree) == 1:
                tree.bind("<Button-1>", lambda event, d = data: self.OnClick(event, d))
        else:
            tree.heading("#0", text = 'No variable selected for preview.', anchor = tk.CENTER)

    def check_quit(self):
        """ Saves the information and closes the window """
        self.newWindow.destroy()

    def OnClick(self,event,d):
        """ On data variable click, updates variable display. """
        if self.tree[0].identify_column(event.x) == '#1':
            treerow = self.tree[0].identify_row(event.y)
            if len(self.tree[0].item(treerow,"values")) > 1:
                data = d[self.tree[0].item(treerow,"values")[0]]
                # Reset treeview
                self.resetTree(self.tree[1])
                # refill second treeview
                self.tree[1]['columns'] = list(d[self.tree[0].item(treerow,"values")[0]].columns)
                # Format columns
                self.tree[1].column("#0", width = self.controller.pages_font.measure('0'*3)+20, 
                        minwidth = 0, stretch=tk.NO)
                for n, cl in enumerate(self.tree[1]['columns']):
                    self.tree[1].column(
                        cl, width = int(
                            self.controller.pages_font.measure(str(cl)*5))+20, 
                            anchor = tk.CENTER, 
                            stretch=tk.NO)
                # Headings
                self.tree[1].heading("#0", text = '', anchor = tk.CENTER)
                for cl in self.tree[1]['columns']:
                    self.tree[1].heading(cl, text = cl, anchor = tk.CENTER)
                # Data filling
                for n,row in enumerate(data.values):
                    if n%2 == 0:
                        self.tree[1].insert(parent = '', index = 'end', iid = n, text = n, 
                                         values = tuple(np.around(row, 3)), tags = ('even',))
                    else:
                        self.tree[1].insert(parent = '', index = 'end', iid = n, text = n, 
                                         values = tuple(np.around(row, 3)), tags = ('odd',))
                self.tree[1].event_generate("<<ThemeChanged>>")
            else:
                self.resetTree(self.tree[1])
                self.tree[1].column("#0", width = self.controller.pages_font.measure('No variable selected for preview.')+20, 
                        minwidth = 0, stretch=tk.NO)
                self.tree[1].heading("#0", text = 'No variable selected for preview.', anchor = tk.CENTER)

    def resetTree(self,tree):
        tree.delete(*tree.get_children())
        tree["columns"] = ()


from vai_lab.Data.Data_core import Data
from vai_lab._import_helper import import_plugin_absolute
from vai_lab._types import PluginSpecsInterface, DataInterface

class InputData(Data):
    def __init__(self):
        super().__init__()
        self.node_name = None
        self.plugin_name = None
        self.output_data = None

    def set_data_in(self, data_in: DataInterface) -> None:
        """Pass existing data from another module to be stored in this class"""
        self._data_in = data_in

    def load_data_from_file(self, filename: str, data_id: str) -> None:
        """Load data from file. Calls parent class method to store data in self.data"""
        super().import_data(filename, data_id)

    def set_options(self, module_config: dict) -> None:
        """Send configuration arguments to plugin

        :param module_config: dict of settings to configure the plugin
        """
        self._module_config = module_config
        self._load_plugin(self._module_config["plugin"]["plugin_name"])

    def set_avail_plugins(self, avail_plugins: PluginSpecsInterface) -> None:
        self._avail_plugins = avail_plugins

    def _load_plugin(self, plugin_name: str) -> None:
        avail_plugins = self._avail_plugins.find_from_readable_name(
            plugin_name)
        self._plugin_name = plugin_name
        self._plugin: PluginSpecsInterface = import_plugin_absolute(globals(),
                                                    avail_plugins["_PLUGIN_PACKAGE"],
                                                    avail_plugins["_PLUGIN_CLASS_NAME"])\
                                                    .__call__()

    def get_result(self):
        return self._data_in



from vai_lab._plugin_templates import UI

class UserInterfaceClass(UI):
    """UserInterfaceClass directly duplicates:
        vai_lab.UserInteraction.User_Interaction_template
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


# -*- coding: utf-8 -*-
class Modelling(object):
    def __init__(self):
        self.node_name = None
        self.plugin_name = None
        self.output_data = None

    def set_input_data(self, data):
        self.input_data = data

    def set_target_data(self, data):
        self.target_data = data

    def _load_plugin(self):
        pass

    def set_plugin_name(self, plugin_name):
        self.plugin_name = plugin_name
        self._load_plugin()

    def set_options(self, options):
        self.options = options

    def solve(self):
        # send the input data to the modelling plugin and assign it to an output property
        pass

    def get_result(self):
        return self.output_data



from typing import List, Iterator, Union
from vai_lab._types import DictT

import ast
import os


class PluginSpecs(ast.NodeVisitor):
    def __init__(self) -> None:
        self.module_dirs: DictT = {}
        self.available_plugins: DictT = {}
        self._get_plugin_files()
        self._get_plugin_specs()

    def _get_plugin_files(self, root_dir: str = None) -> None:
        if root_dir is None:
            root_dir = os.path.dirname(os.path.realpath(__file__))

        for mod_folder in self._get_clean_module_dirs(root_dir):
            mod_path = os.path.join(root_dir, mod_folder, "plugins")
            plugins = self._get_clean_module_dirs(mod_path)

            self.module_dirs[mod_folder] = {}
            self.module_dirs[mod_folder]["root"] = mod_path
            self.module_dirs[mod_folder]["files"] = [
                f for f in plugins if f.endswith(".py")]

    def _get_clean_module_dirs(self, root_dir: str) -> List[str]:
        excludes = ["__init__.py", "__pycache__", "resources"]
        if os.path.exists(root_dir):
            mod_dirs = os.listdir(root_dir)
            for ex in excludes:
                if ex in mod_dirs:
                    mod_dirs.remove(ex)
            return mod_dirs
        else:
            return []

    def _plugins_generator(self) -> Iterator:
        plugin_names = self.module_dirs[self.curr_module]["files"]
        for plugin in plugin_names:
            self.curr_plugin = plugin
            yield os.path.join(self.module_dirs[self.curr_module]["root"], plugin)

    def _modules_generator(self):
        for self.curr_module in self.module_dirs.keys():
            self.available_plugins[self.curr_module] = {}
            plugins = self._plugins_generator()
            for plugin_dir in plugins:
                yield plugin_dir

    def _get_module_package(self):
        if __package__:
            parent_package = __package__.split(".")[0]
        else:
            parent_package = "vai_lab"
        scipt_filename = self.curr_plugin.replace(".py", "")
        return "{0}.{1}.plugins.{2}".format(parent_package, self.curr_module, scipt_filename)

    def _get_plugin_specs(self):
        self.modules = self._modules_generator()
        for plugin_dir in self.modules:
            self.available_plugins[self.curr_module][self.curr_plugin] \
                = {"_PLUGIN_DIR": plugin_dir,
                   "_PLUGIN_PACKAGE": self._get_module_package()
                   }
            with open(plugin_dir, "r") as source:
                tree = ast.parse(source.read())
            self.visit(tree)

    def visit_ClassDef(self, node):
        """Append the name of the main class of the plugin to specs
        Called as a callback from "self.visit(tree)"
        Adds only the first class name - works for now, but may be suboptimal

        :param node: ast node of class name
        """
        c_tag = "_PLUGIN_CLASS_NAME"
        d_tag = "_PLUGIN_CLASS_DESCRIPTION"
        if c_tag not in self.available_plugins[self.curr_module][self.curr_plugin].keys():
            self.available_plugins[self.curr_module][self.curr_plugin][c_tag] = node.name
            ds = ast.get_docstring(node)
            self.available_plugins[self.curr_module][self.curr_plugin][d_tag] = ds

    def visit_Assign(self, node):
        """Append the name of all plugin specs to available_plugins dict
        Plugin specs should be declared with consants beginning with "_PLUGIN_<option_name>"
        Called as a callback from "self.visit(tree)"

        :param node: ast node of class name
        """
        if type(node.targets[0]) == ast.Name:
            if "_PLUGIN" in node.targets[0].id:
                key = node.targets[0].id
                val = eval(compile(ast.Expression(node.value),
                           "<ast expression>", "eval"))
                self.available_plugins[self.curr_module][self.curr_plugin][key] = val

    def _get_option_specs(self, option: str) -> dict:
        """Gets either required OR optional settings for each plugin.

        :param option: str containing either:
                        "_PLUGIN_OPTIONAL_SETTINGS"
                        OR
                        "_PLUGIN_REQUIRED_SETTINGS"

        :returns output: nested dict of options by [Module][Plugin Class Name]
        """
        output : DictT= {}
        for module in self.available_plugins.keys():
            output[module] = {}
            for plugin in self.available_plugins[module].keys():
                if option in self.available_plugins[module][plugin]:
                    cn = self.available_plugins[module][plugin]["_PLUGIN_CLASS_NAME"]
                    output[module][cn] = self.available_plugins[module][plugin][option]
        return output

    def _find_plugin_by_tag_and_value(self, tag: str, name: str) -> Union[DictT,None]:
        for module in self.available_plugins.keys():
            for plugin in self.available_plugins[module].keys():
                if tag in self.available_plugins[module][plugin]:
                    if name in self.available_plugins[module][plugin][tag]:
                        return self.available_plugins[module][plugin]
        return None

    @property
    def names(self):
        return self._get_option_specs("_PLUGIN_READABLE_NAMES")

    @property
    def class_names(self):
        return self._get_option_specs("_PLUGIN_CLASS_NAME")

    @property
    def module_options(self):
        return self._get_option_specs("_PLUGIN_MODULE_OPTIONS")

    @property
    def required_settings(self):
        return self._get_option_specs("_PLUGIN_REQUIRED_SETTINGS")

    @property
    def class_descriptions(self):
        return self._get_option_specs("_PLUGIN_CLASS_DESCRIPTION")

    @property
    def optional_settings(self):
        return self._get_option_specs("_PLUGIN_OPTIONAL_SETTINGS")

    @property
    def available_plugin_names(self):
        names = self.names
        output = []
        for n in names.keys():
            for plugin in names[n]:
                for variant in names[n][plugin].keys():
                    if names[n][plugin][variant] == "default":
                        output.append(variant)
        return output

    def find_from_class_name(self, value):
        return self._find_plugin_by_tag_and_value("_PLUGIN_CLASS_NAME", value)

    def find_from_readable_name(self, value):
        return self._find_plugin_by_tag_and_value("_PLUGIN_READABLE_NAMES", value)

    def print(self, value):
        from pprint import PrettyPrinter
        pp = PrettyPrinter(sort_dicts=False, width=100)
        pp.pprint(value)


if __name__ == "__main__":
    ps = PluginSpecs()
    ps.print(ps.available_plugin_names)
    # ps.print(ps.names)
    # ps.print(ps.class_descriptions)
    # print(list(ps.class_descriptions()['GUI'].values()))




from typing import List
from vai_lab._plugin_templates import UI

import tkinter as tk
from tkinter import messagebox, ttk
from tkinter.filedialog import asksaveasfile, askopenfile, askopenfilename

import os
import numpy as np
import pandas as pd
import math

_PLUGIN_READABLE_NAMES = {"canvas": "default",
                          "state-action": "alias", "robot": "alias"}    # type:ignore
_PLUGIN_MODULE_OPTIONS = {"layer_priority": 2,
                          "required_children": None}                    # type:ignore
_PLUGIN_REQUIRED_SETTINGS = {}                                          # type:ignore
_PLUGIN_OPTIONAL_SETTINGS = {}                                          # type:ignore
_PLUGIN_REQUIRED_DATA = {"X"}                                           # type:ignore


class CanvasInput(tk.Frame, UI):
    """Method of user interaction for state-action pairs"""

    def __init__(self, parent, controller, config: dict):
        self.parent = parent
        super().__init__(self.parent, bg=self.parent['bg'])
        self.controller = controller

        self.controller.title('Canvas Input')
        self.notebook = ttk.Notebook(self)
        self.notebook.grid(row=0, column=0, columnspan=7,
                           rowspan=12, pady=15)
        self.frame: List[tk.Frame] = []
        self.tree: List[ttk.Treeview] = []
        self.canvas: List[tk.Canvas] = []
        self.draw: List[tk.StringVar] = []
        self.state: List[tk.StringVar] = []
        self.type: List[str] = []
        self.clock: List[tk.StringVar] = []
        self.save_path = ''
        self.saved = True

    def set_data_in(self, data_in):
        req_check = [
            r for r in _PLUGIN_REQUIRED_DATA if r not in data_in.keys()]
        if len(req_check) > 0:
            raise Exception("Minimal Data Requirements not met"
                            + "\n\t{0} ".format("CanvasInput")
                            + "requires data: {0}".format(_PLUGIN_REQUIRED_DATA)
                            + "\n\tThe following data is missing:"
                            + "\n\t\u2022 {}".format(",\n\t\u2022 ".join([*req_check])))
        self._data_in = data_in
        self._load_classes_from_data()

    # Causes static type error as it overloads the method in the parent class
    def configure(self, config: dict):          # type:ignore
        self._config = config

    def class_list(self):
        """Getter for required _class_list variable

        :return: list of class labels
        :rtype: list of strings
        """
        return self._class_list

    def _load_classes_from_data(self):
        """Setter for required _class_list variable

        :param value: labels for state-action pair headers
        :type value: list of strings
        """

        self._class_list = list(self._data_in["X"].columns)
        self._class_list = np.squeeze(self._class_list)
        islist = []
        for ii in np.arange(len(self._class_list)):
            islist.append(isinstance(self._class_list[ii], list))
        if not all(islist):
            self._class_list = [list(self._class_list)]

        self.out_data = {}
        self.out_data[0] = {}
        for ii in np.arange(len(self._class_list)):
            self.out_data[ii] = {}
            for k in np.arange(len(self._class_list[ii])):
                self._class_list[ii][k] = self._class_list[ii][k].replace(
                    ' ', '_').lower()
                self.out_data[ii][self._class_list[ii][k]] = []

        for ii in np.arange(len(self.out_data)):
            self.frame.append(tk.Frame(self, bg=self.parent['bg']))
            self.frame[-1].grid(row=0, column=0, sticky="nsew", pady=15)
            # if self._class_list[ii][0].split('_')[1] == 'a':
            if self._class_list[ii][0].split('_')[1] == 'a':
                self.type.append('Rotating')
                self.clock.append(tk.StringVar())
                self.clock[ii].set('clock')
            else:
                self.type.append('Sliding')
                self.clock.append(-1)
            self.notebook.add(
                self.frame[-1],
                text='Object ' + str(ii) + ' - ' + self.type[-1])

            # Create a canvas widget
            self.width, self.height = 600, 600
            self.canvas.append(tk.Canvas(
                self.frame[-1], width=self.width,
                height=self.height, background="white"))
            self.canvas[-1].grid(row=0, column=0, columnspan=4, rowspan=2)
            self.canvas[-1].bind('<Button-1>', self.draw_dot)
            self.canvas[-1].bind("<B1-Motion>", self.on_drag)
            # self.canvas[-1].bind('<Motion>', self.motion)

            # Booleans to identify what the user wants to do
            self.draw.append(tk.StringVar())
            self.draw[ii].set('draw')
            self.state.append(tk.StringVar())
            self.state[ii].set('state')
            # Buttons under the canvas
            self.button_draw = tk.Radiobutton(
                self.frame[-1], text='Draw', fg='white', bg=self.parent['bg'],
                height=3, width=20, var=self.draw[ii],
                selectcolor='black', value='draw').grid(column=4, row=3)
            self.button_drag = tk.Radiobutton(
                self.frame[-1], text='Move', fg='white', bg=self.parent['bg'],
                height=3, width=20, var=self.draw[ii],
                selectcolor='black', value='drag').grid(column=5, row=3)
            self.clock_arc = tk.Radiobutton(
                self.frame[-1], text='Clockwise', fg='white', bg=self.parent['bg'],
                height=3, width=20, var=self.clock[ii],
                selectcolor='black', value='clock').grid(column=6, row=3)
            self.countclock_arc = tk.Radiobutton(
                self.frame[-1], text='Counterclockwise', fg='white', bg=self.parent['bg'],
                height=3, width=20, var=self.clock[ii],
                selectcolor='black', value='countclock').grid(column=7, row=3)
            # self.button_edit = tk.Radiobutton(
            #     self.frame[-1], text = 'Edit', fg = 'white', bg = self.parent['bg'],
            #     height = 3, width = 20, var = self.draw[ii],
            #     selectcolor = 'black', value = 'edit').grid(column = 6,row = 3)
            self.button_save = tk.Button(
                self.frame[-1], text='Save', fg='white', bg=self.parent['bg'],
                height=3, width=20,
                command=self.save_file).grid(column=1, row=3)
            self.button_upload = tk.Button(
                self.frame[-1], text='Upload coordinates', fg='white',
                bg=self.parent['bg'], height=3, width=20,
                command=self.upload_sa).grid(column=0, row=3)
            self.button_reset = tk.Button(
                self.frame[-1], text='Reset', fg='white',
                bg=self.parent['bg'], height=3, width=20,
                command=self.reset).grid(column=2, row=3)
            button_main = tk.Button(
                self.frame[-1], text="Done", fg='white',
                bg=self.parent['bg'], height=3, width=20,
                command=self.check_quit).grid(column=3, row=3)
            style = ttk.Style()
            style.configure(
                "Treeview", background='white', foreground='white',
                rowheight=25, fieldbackground='white',
                font=self.controller.pages_font)
            style.configure("Treeview.Heading",
                            font=self.controller.pages_font)
            style.map('Treeview', background=[('selected', 'grey')])

            tree_frame = tk.Frame(self.frame[-1])
            tree_frame.grid(row=0, column=4, columnspan=3, rowspan=10)

            tree_scrollx = tk.Scrollbar(tree_frame, orient='horizontal')
            tree_scrollx.pack(side=tk.BOTTOM, fill=tk.X)
            tree_scrolly = tk.Scrollbar(tree_frame)
            tree_scrolly.pack(side=tk.RIGHT, fill=tk.Y)

            self.tree.append(ttk.Treeview(
                tree_frame,
                yscrollcommand=tree_scrolly.set,
                xscrollcommand=tree_scrollx.set))
            self.tree[-1].pack()

            tree_scrollx.config(command=self.tree[-1].xview)
            tree_scrolly.config(command=self.tree[-1].yview)

            if self.type[-1] == 'Rotating':
                self.x_ini, self.y_ini = int(self.width/2), int(self.height/2)
                self.r = int(np.min([self.x_ini, self.y_ini])/1.5)
                self.create_circle(self.x_ini, self.y_ini, 4,
                                   fill="#DDD", outline="", width=4)
                self.create_circle(self.x_ini, self.y_ini, self.r,
                                   fill="", outline="#DDD", width=4)

            self.tree[-1]['columns'] = self._class_list[ii]

            # Format columns
            self.tree[-1].column("#0", width=50)
            for n, cl in enumerate(self._class_list[ii]):
                self.tree[-1].column(
                    cl, width=int(
                        self.controller.pages_font.measure(str(cl)))+20,
                    minwidth=50, anchor=tk.CENTER)

            # Headings
            self.tree[-1].heading("#0", text="Sample", anchor=tk.CENTER)
            for cl in self._class_list[ii]:
                self.tree[-1].heading(cl, text=cl, anchor=tk.CENTER)

            # for nn in np.arange(N):
            #     self.tree[-1].insert(
            #                 parent = '', index = 'end', iid = (nn+1)*1000, text = 'Trial '+str(nn),
            #                 values = tuple())

            self.tree[-1].tag_configure('odd', foreground='black',
                                        background='#E8E8E8')
            self.tree[-1].tag_configure('even', foreground='black',
                                        background='#DFDFDF')

            # Define double-click on row action
            self.tree[-1].bind("<Double-1>", self.OnDoubleClick)

    def draw_dot(self, event):

        ii = self.notebook.index(self.notebook.select())
        if self.draw[ii].get() == 'drag':
            self.selected = self.canvas[ii].find_overlapping(
                event.x-10, event.y-10, event.x+10, event.y+10)
            if self.selected:
                if len(self.selected) >= 2:
                    self.canvas[ii].selected = self.selected[-2]
                else:
                    self.canvas[ii].selected = self.selected[-1]
                self.canvas[ii].startxy = (event.x, event.y)
            else:
                self.canvas[ii].selected = None

        if self.draw[ii].get() == 'draw':
            # Draw an oval in the given coordinates
            self.saved = False
            # Angle of the selected point with respect to the circle center
            if self.type[ii] == 'Rotating':
                alpha, x_o, y_o = self.angle_calc(event.x, event.y,
                                                  circxy=True)
                if self.state[ii].get() == 'state':  # Write state coordinates
                    self.canvas[ii].create_oval(
                        x_o-3, y_o-3, x_o+3, y_o+3, fill="black", width=0,
                        tags=("state"+str(ii)+'-'
                              + str(len(self.out_data[ii]['state_a']))))
                    self.out_data[ii]['state_a'].append((
                        -np.rad2deg(alpha)+360) % 360)
                    self.state[ii].set('action')
                    # Update coordinates in corresponding row if exists.
                    if self.tree[ii].selection():
                        n = int(self.tree[ii].selection()[0]) + 1
                        if n % 2 == 0:
                            tag = ('even',)
                        else:
                            tag = ('odd',)
                        self.tree[ii].insert(
                            parent='', index='end', iid=n, text=n+1,
                            values=tuple(self.dict2mat(
                                self.out_data[ii])[n, :].astype(int)),
                            tags=tag)
                        self.tree[ii].selection_set(str(n))
                    else:
                        self.tree[ii].insert(
                            parent='', index='end', iid=0, text=1,
                            values=tuple(self.dict2mat(
                                self.out_data[ii])[0, :].astype(int)),
                            tags=('even',))
                        self.tree[ii].selection_set(str(0))

                elif self.state[ii].get() == 'action':
                    if self.clock[ii].get() == 'clock':
                        start = self.out_data[ii]['state_a'][-1]
                        end = np.rad2deg(-alpha)
                    else:
                        start = self.out_data[ii]['state_a'][-1] - 360
                        end = 360 - np.rad2deg(alpha)
                    self.create_circle_arc(
                        self.x_ini, self.y_ini, self.r, fill="",
                        outline="red", start=start, end=end, width=2,
                        style=tk.ARC, tags=("action"+str(ii)+'-' + str(len(
                            self.out_data[ii]['action_a']))))
                    self.out_data[ii]['action_a'].append((
                        -np.rad2deg(alpha)+360) % 360)
                    self.state[ii].set('state')
                    n = int(self.tree[ii].selection()[0])
                    self.tree[ii].item(
                        self.tree[ii].get_children()[n], text=n+1,
                        values=tuple(self.dict2mat(
                            self.out_data[ii])[n, :].astype(int)))
            else:
                if self.state[ii].get() == 'state':  # Write state coordinates
                    self.canvas[ii].create_oval(
                        event.x-3, event.y-3, event.x+3, event.y+3,
                        fill="black", width=0,
                        tags=("state"+str(ii)+'-' + str(len(
                            self.out_data[ii]['state_x']))))
                    self.out_data[ii]['state_x'].append(event.x)
                    self.out_data[ii]['state_y'].append(event.y)
                    self.state[ii].set('action')
                    # Update coordinates in corresponding row if exists.
                    if self.tree[ii].selection():
                        n = int(self.tree[ii].selection()[0]) + 1
                        if n % 2 == 0:
                            tag = ('even',)
                        else:
                            tag = ('odd',)
                        self.tree[ii].insert(
                            parent='', index='end', iid=n, text=n+1,
                            values=tuple(self.dict2mat(
                                self.out_data[ii])[n, :].astype(int)),
                            tags=tag)
                        self.tree[ii].selection_set(str(n))
                    else:
                        self.tree[ii].insert(
                            parent='', index='end', iid=0, text=1,
                            values=tuple(self.dict2mat(
                                self.out_data[ii])[0, :].astype(int)),
                            tags=('even',))
                        self.tree[ii].selection_set(str(0))

                elif self.state[ii].get() == 'action':
                    self.canvas[ii].create_line(
                        self.out_data[ii]['state_x'][-1],
                        self.out_data[ii]['state_y'][-1], event.x, event.y,
                        fill="red", arrow=tk.LAST,
                        tags=("action"+str(ii)+'-' + str(len(
                            self.out_data[ii]['action_x']))))
                    self.out_data[ii]['action_x'].append(event.x)
                    self.out_data[ii]['action_y'].append(event.y)
                    self.state[ii].set('state')
                    n = int(self.tree[ii].selection()[0])
                    self.tree[ii].item(
                        self.tree[ii].get_children()[n], text=n+1,
                        values=tuple(self.dict2mat(
                            self.out_data[ii])[n, :].astype(int)))

    def on_drag(self, event):

        ii = self.notebook.index(self.notebook.select())
        if self.draw[ii].get() == 'drag' and self.canvas[ii].selected:
            # move the selected item
            n = int(self.canvas[ii].gettags("current")[0].split('-')[1])

            if self.type[ii] == 'Rotating':
                alpha, dx, dy = self.angle_calc(event.x, event.y, circxy=True)
                _, dx_pr, dy_pr = self.angle_calc(
                    self.canvas[ii].startxy[0], self.canvas[ii].startxy[1],
                    circxy=True)
                xa = self.r * np.cos(
                    np.deg2rad(self.out_data[ii]['action_a'][n])) + self.x_ini
                ya = self.r * np.sin(
                    np.deg2rad(self.out_data[ii]['action_a'][n])) + self.y_ini
                if self.canvas[ii].gettags(
                        "current")[0].split('-')[0] == 'state'+str(ii):  # If state, updates both the state placement and the arrow
                    self.canvas[ii].move(
                        self.canvas[ii].selected, dx-dx_pr,
                        dy-dy_pr)
                    self.out_data[ii]['state_a'][n] = (
                        360-np.rad2deg(alpha)) % 360

                    if self.clock[ii].get() == 'clock':
                        start = self.out_data[ii]['action_a'][n]
                        end = np.rad2deg(-alpha)
                    else:
                        start = self.out_data[ii]['action_a'][n] - 360
                        end = 360 - np.rad2deg(alpha)

                    self.canvas[ii].itemconfigure(
                        "action"+str(ii)+'-'+str(n), start=start,
                        extent=end - start)

                elif self.canvas[ii].gettags(
                        "current")[0].split('-')[0] == 'action'+str(ii):  # If action, updates the arrow end of the line position

                    self.out_data[ii]['action_a'][n] = (
                        360-np.rad2deg(alpha)) % 360

                    if self.clock[ii].get() == 'clock':
                        start = self.out_data[ii]['state_a'][n]
                        end = np.rad2deg(-alpha)
                    else:
                        start = self.out_data[ii]['state_a'][n] - 360
                        end = 360 - np.rad2deg(alpha)

                    self.canvas[ii].itemconfigure(
                        "action"+str(ii)+'-'+str(n), start=start,
                        extent=end - start)

            else:
                # calculate distance moved from last position
                dx = event.x - self.canvas[ii].startxy[0]
                dy = event.y - self.canvas[ii].startxy[1]
                if self.canvas[ii].gettags(
                        "current")[0].split('-')[0] == 'state'+str(ii):  # If state, updates both the state placement and the arrow
                    self.canvas[ii].move(self.canvas[ii].selected, dx, dy)
                    self.canvas[ii].coords(
                        "action"+str(ii)+'-'+str(n),
                        (event.x, event.y, self.out_data[ii]['action_x'][n],
                         self.out_data[ii]['action_y'][n]))
                    self.out_data[ii]['state_x'][n] = event.x
                    self.out_data[ii]['state_y'][n] = event.y
                elif self.canvas[ii].gettags(
                        "current")[0].split('-')[0] == 'action'+str(ii):  # If action, updates the arrow end of the line position
                    self.canvas[ii].coords(
                        self.canvas[ii].gettags("current")[0],
                        (self.out_data[ii]['state_x'][n],
                         self.out_data[ii]['state_y'][n], event.x, event.y))
                    self.out_data[ii]['action_x'][n] = event.x
                    self.out_data[ii]['action_y'][n] = event.y
            # update last position
            self.canvas[ii].startxy = (event.x, event.y)
            self.tree[ii].item(self.tree[ii].get_children()[n], text=n+1,
                               values=tuple(self.dict2mat(
                                   self.out_data[ii])[n, :].astype(int)))
            self.tree[ii].selection_set(str(n))

    def checkered(self, line_distance):

        ii = self.notebook.index(self.notebook.select())
        # vertical lines at an interval of "line_distance" pixel
        for x in range(line_distance, self.width, line_distance):
            self.canvas[ii].create_line(x, 0, x, self.height, fill="#476042")
        # horizontal lines at an interval of "line_distance" pixel
        for y in range(line_distance, self.height, line_distance):
            self.canvas[ii].create_line(0, y, self.width, y, fill="#476042")

    def check_quit(self):
        if not self.saved:
            response = messagebox.askokcancel(
                "Exit?",
                "Do you want to leave the program without saving?")
            if response:
                self.controller.destroy()
        else:
            response = messagebox.askokcancel(
                "Exit?",
                "Are you sure you are finished?")
            self.controller.destroy()

    def save_file_as(self):
        self.save_path = asksaveasfile(mode='w')
        self.save_file()

    def save_file(self):

        if self.save_path == '':
            self.save_path = asksaveasfile(defaultextension='.txt', filetypes=[('Text file', '.txt'),
                                                                               ('CSV file',
                                                                                '.csv'),
                                                                               ('All Files', '*.*')])
        # asksaveasfile return `None` if dialog closed with "cancel".
        if self.save_path is not None:
            # Transform input for output file
            data = {}
            samples = np.max([len(self.out_data[element][
                list(self.out_data[element].keys())[0]])
                for element in self.out_data])
            for ii in self.out_data.keys():
                for c in self.out_data.keys():
                    a, b = c.split('_')
                    aux = self.out_data[c]
                    aux.extend([-1]*(samples - len(self.out_data[c])))
                    data[a+str(ii)+'_'+b] = aux

            filedata = pd.DataFrame(data, columns=data.keys()).to_string()
            self.save_path.seek(0)  # Move to the first row to overwrite it
            self.save_path.write(filedata)
            self.save_path.flush()  # Save without closing
            # typically the above line would do. however this is used to ensure that the file is written
            os.fsync(self.save_path.fileno())
            self.saved = True

    def on_return(self, event):
        ii = self.notebook.index(self.notebook.select())
        val = self.tree[ii].item(self.treerow)['values']
        val = [float(i) for i in val]
        val[int(self.treecol[1:])-1] = float(self.entry.get())
        self.tree[ii].item(self.treerow, values=val)
        self.entry.destroy()
        self.saved = False

        if self.type[ii] == 'Rotating':
            dx, dy = self.coord_calc(np.deg2rad(360 - val[0]))
            dx_pr, dy_pr = self.coord_calc(np.deg2rad(
                360 - self.out_data[ii]['state_a'][self.treerow]))
            self.out_data[ii]['state_a'][self.treerow] = float(val[0])
            self.canvas[ii].move("state"+str(ii)+'-'+str(self.treerow), dx-dx_pr,
                                 dy-dy_pr)

            self.out_data[ii]['action_a'][self.treerow] = float(val[1])
            if self.clock[ii].get() == 'clock':
                start = self.out_data[ii]['state_a'][self.treerow]
                end = self.out_data[ii]['action_a'][self.treerow]
            else:
                start = self.out_data[ii]['state_a'][self.treerow] - 360
                end = self.out_data[ii]['action_a'][self.treerow]

            self.canvas[ii].itemconfigure(
                "action"+str(ii)+'-'+str(self.treerow), start=start,
                extent=end - start)
        else:
            # calculate distance moved from last position
            dx_s = val[0] - self.out_data[ii]['state_x'][self.treerow]
            dy_s = val[1] - self.out_data[ii]['state_y'][self.treerow]

            self.canvas[ii].move("state"+str(ii)+'-' +
                                 str(self.treerow), dx_s, dy_s)

            self.out_data[ii]['state_x'][self.treerow] = val[0]
            self.out_data[ii]['state_y'][self.treerow] = val[1]

            self.canvas[ii].coords(
                "action"+str(ii)+'-'+str(self.treerow),
                (self.out_data[ii]['state_x'][self.treerow],
                 self.out_data[ii]['state_y'][self.treerow], val[2], val[3]))
            self.out_data[ii]['action_x'][self.treerow] = val[2]
            self.out_data[ii]['action_y'][self.treerow] = val[3]

    def OnDoubleClick(self, event):
        """ Executed when a row is double clicked.
        Opens an entry box to edit a cell and updates the canvas and the 
        stored data. """

        ii = self.notebook.index(self.notebook.select())
        self.treerow = int(self.tree[ii].identify_row(event.y))
        self.treecol = self.tree[ii].identify_column(event.x)

        # get column position info
        x, y, width, height = self.tree[ii].bbox(self.treerow, self.treecol)

        # y-axis offset
        pady = height // 2
        # pady = 0

        if hasattr(self, 'entry'):
            self.entry.destroy()

        self.entry = tk.Entry(self.tree[ii], justify='center')

        if int(self.treecol[1:]) > 0:
            self.entry.insert(
                0, self.tree[ii].item(self.treerow)['values'][int(str(self.treecol[1:]))-1])
            # self.entry['selectbackground'] = '#123456'
            self.entry['exportselection'] = False

            self.entry.focus_force()
            self.entry.bind("<Return>", self.on_return)
            self.entry.bind("<Escape>", lambda *ignore: self.entry.destroy())

            self.entry.place(x=x,
                             y=y + pady,
                             anchor=tk.W, width=width)

    def upload_sa(self):

        ii = self.notebook.index(self.notebook.select())
        filename = askopenfilename(initialdir=os.getcwd(),
                                   title='Select a file',
                                   defaultextension='.txt',
                                   filetypes=[('Text file', '.txt'),
                                              ('CSS file', '.css'),
                                              ('All Files', '*.*')])
        if filename is not None:
            data = open(filename, 'r')
            read = False
            self.draw[ii].set('drag')
            for n, point in enumerate(data):
                if read:  # Not elegant at all, just to omit the header.
                    i, sx, sy, ax, ay = point.split()
                    # Draw an oval in the given coordinates
                    self.canvas[ii].create_oval(
                        float(sx)-3, float(sy)-3, float(sx)+3, float(sy)+3,
                        fill="black", width=0,
                        tags=("state-" + str(len(
                            self.out_data[ii]['state_x']))))
                    self.canvas[ii].create_line(
                        float(sx), float(sy), float(ax), float(ay),
                        fill="red", arrow=tk.LAST,
                        tags=("action"+str(ii)+"-" + str(len(
                            self.out_data[ii]['action_x']))))
                    self.out_data[ii]['state_x'].append(sx)
                    self.out_data[ii]['state_y'].append(sy)
                    self.out_data[ii]['action_x'].append(ax)
                    self.out_data[ii]['action_y'].append(ay)
                    self.tree.insert(
                        parent='', index='end',
                        iid=len(self.out_data[ii]['action_x'])-1,
                        text=len(self.out_data[ii]['action_x']),
                        values=tuple(self.dict2mat(
                            self.out_data[ii])
                            [len(self.out_data[ii]
                                 ['action_x'])-1, :].astype(int)))
                    self.tree.selection_set(str(len(
                        self.out_data[ii]['action_x'])-1))
                else:
                    read = True

    def reset(self):

        ii = self.notebook.index(self.notebook.select())
        msg = messagebox.askyesnocancel(
            'Info', 'Are you sure you want to reset the canvas?')
        if msg:
            self.canvas[ii].delete(tk.ALL)  # Reset canvas
            self.state[ii].set('state')
            self.out_data = {}
            for key in self._class_list[ii]:
                self.out_data[key] = []

            for record in self.tree[ii].get_children():  # Reset treeview
                self.tree[ii].delete(record)
            if self.type[ii] == 'Rotating':
                self.create_circle(self.x_ini, self.y_ini, 4,
                                   fill="#DDD", outline="", width=4)
                self.create_circle(self.x_ini, self.y_ini, self.r,
                                   fill="", outline="#DDD", width=4)

    def dict2mat(self, X):

        max_size = 0
        for key in X:
            if len(X[key]) > max_size:
                max_size = len(X[key])

        result = -np.ones((max_size, len(X)))
        for k, key in enumerate(X):
            result[:len(X[key]), k] = X[key]

        return result

    def angle_calc(self, x, y, circxy=False):
        if (x == self.x_ini) & (y < self.y_ini):
            alpha = math.pi/2
        elif (x == self.x_ini) & (y > self.y_ini):
            alpha = 3*math.pi/2
        else:
            alpha = np.arctan((y-self.y_ini)/(x-self.x_ini))
            alpha += ((x-self.x_ini) < 0) * math.pi
            alpha += (alpha < 0) * 2*math.pi
        if circxy:
            cx, cy = self.coord_calc(alpha)
            return alpha, cx, cy
        else:
            return alpha

    def coord_calc(self, alpha):
        cx = self.r * np.cos(alpha) + self.x_ini
        cy = self.r * np.sin(alpha) + self.y_ini
        return cx, cy

    def create_circle(self, x, y, r, **kwargs):

        return self.canvas[self.notebook.index(
            self.notebook.select())].create_oval(x-r, y-r, x+r, y+r, **kwargs)

    def create_circle_arc(self, x, y, r, **kwargs):

        if "start" in kwargs and "end" in kwargs:
            kwargs["extent"] = kwargs["end"] - kwargs["start"]
            # if self.clock == 'countclock':
            #     kwargs["extent"] += (kwargs["extent"] < 0) * 360
            del kwargs["end"]
        return self.canvas[self.notebook.index(
            self.notebook.select())].create_arc(x-r, y-r, x+r, y+r, **kwargs)



# -*- coding: utf-8 -*-
from vai_lab.GUI.GUI_core import GUI
class UserInteraction(GUI):
    """UserFeedback core class directly uses GUI_core.py"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



from vai_lab._import_helper import import_plugin_absolute
from vai_lab._types import PluginSpecsInterface, DataInterface, EnvironmentPluginInterface

class Environment(object):
    def __init__(self) -> None:
        self.output_data: DataInterface

    def set_avail_plugins(self, avail_plugins: PluginSpecsInterface) -> None:
        self._avail_plugins = avail_plugins

    def set_data_in(self, data_in: DataInterface) -> None:
        self._data_in = data_in

    def _load_plugin(self, plugin_name: str) -> None:
        avail_plugins = self._avail_plugins.find_from_readable_name(
            plugin_name)
        self._plugin_name = plugin_name
        self._plugin: EnvironmentPluginInterface = import_plugin_absolute(globals(),
                                              avail_plugins["_PLUGIN_PACKAGE"],
                                              avail_plugins["_PLUGIN_CLASS_NAME"])\
            .__call__()

    def set_options(self, module_config: dict) -> None:
        """Send configuration arguments to plugin

        :param module_config: dict of settings to configure the plugin
        """
        self._module_config = module_config
        self._load_plugin(self._module_config["plugin"]["plugin_name"])

    def launch(self) -> None:
        self._plugin.set_data_in(self._data_in)
        self._plugin.configure(self._module_config["plugin"])
        self._plugin.connect()
        self._plugin.load_model()
        self._plugin.run_simulation()
        

    def get_result(self) -> DataInterface:
        # return self.output_data
        return self._data_in



# -*- coding: utf-8 -*-
from vai_lab._import_helper import import_plugin_absolute
class Modelling(object):
    def __init__(self):
        self.output_data = None

    def set_avail_plugins(self,avail_plugins):
        self._avail_plugins = avail_plugins

    def set_data_in(self,data_in):
        self._data_in = data_in

    def _load_plugin(self, plugin_name:str):
        avail_plugins = self._avail_plugins.find_from_readable_name(plugin_name)
        self._plugin_name = plugin_name
        self._plugin = import_plugin_absolute(globals(),\
                                avail_plugins["_PLUGIN_PACKAGE"],\
                                avail_plugins["_PLUGIN_CLASS_NAME"])\
                                .__call__()

    def set_options(self, module_config: dict):
        """Send configuration arguments to plugin

        :param module_config: dict of settings to configure the plugin
        """
        self._module_config = module_config
        self._load_plugin(self._module_config["plugin"]["plugin_name"])

    def launch(self):
        self._plugin.set_data_in(self._data_in)
        self._plugin.configure(self._module_config["plugin"])
        self._plugin.solve()
        self.output_data = self._data_in.copy()
        # self.output_data = self._plugin._test(self.output_data)

    def get_result(self):
        return self.output_data



import tkinter as tk
import os
from PIL import Image, ImageTk

from tkinter.filedialog import askopenfilename, askdirectory
import pandas as pd

from vai_lab.Data.xml_handler import XML_handler
from vai_lab.utils.plugins.dataLoader import dataLoader
from vai_lab._import_helper import get_lib_parent_dir, rel_to_abs

_PLUGIN_READABLE_NAMES = {"main": "default",
                          "main page": "alias",
                          "launch page": "alias"}                               # type:ignore
_PLUGIN_MODULE_OPTIONS = {"layer_priority": 1,
                          "required_children": ['aidCanvas', 'pluginCanvas']}   # type:ignore
_PLUGIN_REQUIRED_SETTINGS = {}                                                  # type:ignore
_PLUGIN_OPTIONAL_SETTINGS = {}                                                  # type:ignore


class MainPage(tk.Frame):
    """Splash and Menu page used as GUI entry-point"""

    def __init__(self, parent, controller, config: dict):
        " Here we define the main frame displayed upon opening the program."
        " This leads to the different methods to provide feedback."
        super().__init__(parent, bg=parent['bg'])
        self.controller = controller
        self.controller.title('AI Assisted Virtual Laboratories')

        self.controller.output_type = 'regression'
        self.out_data = pd.DataFrame()
        self.bg = parent['bg']

        self.script_dir = get_lib_parent_dir()
        self.my_img1 = ImageTk.PhotoImage(
            Image.open(
                os.path.join(
                    self.script_dir,
                    'utils',
                    'resources',
                    'Assets',
                    'VAILabs.png')
            ).resize((350, 350))
        )

        self.grid_columnconfigure(0, weight=1)

        frame1 = tk.Frame(self, bg=self.bg)
        frame2 = tk.Frame(self, bg=self.bg)
        frame3 = tk.Frame(self, bg=self.bg)

        self.my_label = tk.Label(frame1,
                                 image=self.my_img1,
                                 bg=parent['bg'])

        self.my_label.grid(column=0,
                           row=0,
                           rowspan=10,
                           columnspan=4,
                           pady=10,
                           sticky=tk.NE)

        my_label = tk.Label(frame2,
                            text='Indicate the data folder and define the pipeline.',
                            pady=10,
                            font=controller.title_font,
                            bg=parent['bg'],
                            fg='white')
        my_label.grid(column=0,
                      row=11,
                      columnspan=4)

        tk.Button(frame3,
                  text='Upload data matrix',
                  fg='white',
                  font=controller.title_font,
                  bg=parent['bg'],
                  height=3,
                  width=20,
                  command=self.upload_data_folder,
                  ).grid(column=0, row=12)
        tk.Button(frame3,
                  text='Upload data path',
                  fg='white',
                  font=controller.title_font,
                  bg=parent['bg'],
                  height=3,
                  width=20,
                  command=self.upload_data_path,
                  ).grid(column=1, row=12)
        self.controller.Datalabel = tk.Label(frame3,
                                             text='Optional',
                                             pady=10,
                                             padx=10,
                                             font=controller.title_font,
                                             bg=parent['bg'],
                                             fg='white')
        self.controller.Datalabel.grid(column=2,
                                       row=12)

        self.interactButton = tk.Button(frame3,
                                        text='Interact with canvas',
                                        fg='white',
                                        font=controller.title_font,
                                        bg=parent['bg'],
                                        height=3,
                                        width=20,
                                        # state=tk.DISABLED,
                                        command=lambda: self.canvas(
                                            "aidCanvas")
                                        )
        self.interactButton.grid(column=0, row=13)

        self. uploadButton = tk.Button(frame3,
                                       text='Upload XML file',
                                       fg='white',
                                       font=controller.title_font,
                                       bg=parent['bg'],
                                       height=3,
                                       width=20,
                                    #    state=tk.DISABLED,
                                       command=self.upload_xml,
                                       )
        self. uploadButton.grid(column=1, row=13)

        self.controller.XMLlabel = tk.Label(frame3,
                                            text='Incomplete',
                                            pady=10,
                                            padx=10,
                                            font=controller.title_font,
                                            bg=parent['bg'],
                                            fg='white')
        self.controller.XMLlabel.grid(column=2,
                                      row=13)
        self.PluginButton = tk.Button(frame3,
                                      text='Modules plugins',
                                      fg='white',
                                      font=controller.title_font,
                                      bg=parent['bg'],
                                      height=3,
                                      width=20,
                                      state=tk.DISABLED,
                                      command=lambda: self.canvas(
                                          "pluginCanvas"),
                                      )
        self.PluginButton.grid(column=0, row=14)
        self.RunButton = tk.Button(frame3,
                                   text='Run Pipeline',
                                   fg='white',
                                   font=controller.title_font,
                                   bg=parent['bg'],
                                   height=3,
                                   width=20,
                                   state=tk.DISABLED,
                                   command=self.controller.destroy,
                                   )
        self.RunButton.grid(column=1, row=14)

        self.controller.XML = tk.IntVar()
        self.controller.Data = tk.IntVar()
        self.controller.Plugin = tk.IntVar()
        self.controller.XML.set(False)
        self.controller.Data.set(False)
        self.controller.Plugin.set(False)

        self.controller.XML.trace('w', self.trace_XML)
        # self.controller.Data.trace('w', self.trace_Data)
        self.controller.Plugin.trace('w', self.trace_Plugin)

        frame1.grid(column=0, row=0, sticky="n")
        frame2.grid(column=0, row=1, sticky="n")
        frame3.grid(column=0, row=2, sticky="n")

    def set_data_in(self, _):
        pass

    def trace_XML(self, *args):
        """ Checks if XML variable has been updated
        """
        if self.controller.XML.get():
            self.controller.XMLlabel.config(text = 'Done!', fg = 'green')
            # if self.controller.Data.get():
            self.PluginButton.config(state = 'normal')
            if self.controller.Plugin.get():
                self.RunButton.config(state='normal')

    # def trace_Data(self, *args):
    #     """ Checks if Data variable has been updated
    #     """
    #     if self.controller.Data.get():
    #         self.controller.Datalabel.config(text='Done!', fg='green')
    #         self.interactButton.config(state='normal')
    #         self.uploadButton.config(state='normal')

    def trace_Plugin(self, *args):
        """ Checks if Plugin variable has been updated
        """
        if self.controller.Plugin.get():
            self.RunButton.config(state='normal')

    def canvas(self, name: str):
        """ Shows the canvas frame.
        :param name: string type of name of desired canvas.
        """
        self.controller._show_frame(name)

    def upload_xml(self):
        """ Indicates the XML file containng the desired pipeline """
        filename = askopenfilename(initialdir=os.getcwd(),
                                   title='Select a file',
                                   defaultextension='.xml',
                                   filetypes=[('XML file', '.xml'),
                                              ('All Files', '*.*')])
        if filename is not None and len(filename) > 0:
            self.controller._append_to_output("xml_filename", filename)
            self.controller.XML.set(True)

    def upload_data_file(self):
        """ Opens a window to indicate the path to the data. """

        self.newWindow = tk.Toplevel(self.controller)
        # Window options
        self.newWindow.title('Data upload')
        self.tk.call('wm', 'iconphoto', self.newWindow, ImageTk.PhotoImage(
            file=os.path.join(os.path.join(
                self.script_dir,
                'utils',
                'resources',
                'Assets',
                'VAILabsIcon.ico'))))
        # self.newWindow.geometry("600x200")

        frame1 = tk.Frame(self.newWindow)
        self.label_list = []
        self.var = ['X', 'Y', 'X_test', 'Y_test']
        self.filenames = ['']*len(self.var)
        for r, v in enumerate(self.var):
            v = v + '*' if r == 0 else v
            tk.Label(frame1, text=v,
                     pady=10,
                     padx=10,
                     fg='black'
                     ).grid(column=0, row=r)
            tk.Button(frame1,
                      text="Browse",
                      command=lambda a=r: self.upload_file(a)
                      ).grid(column=1, row=r)
            tk.Button(frame1,
                      text="Delete",
                      command=lambda a=r: self.delete_file(a)
                      ).grid(column=2, row=r)
            self.label_list.append(tk.Label(frame1, text=' '*120,
                                            pady=10,
                                            padx=10,
                                            fg='black'
                                            ))
            self.label_list[-1].grid(column=3, row=r)
        frame2 = tk.Frame(self.newWindow)
        tk.Label(frame2, text='* This data file is mandatory.',
                 pady=10,
                 padx=10,
                 fg='black'
                 ).pack(side=tk.LEFT)
        tk.Button(frame2,
                  text="Done",
                  command=self.start_dataloader
                  ).pack(side=tk.RIGHT, padx=(0, 5))
        frame1.grid(column=0, row=0, sticky="nsew")
        frame2.grid(column=0, row=1, sticky="nsew")

    def upload_file(self, r):
        """ Asks for a file and stores the path and displays it.
        :param r: int type of data variable number.
        """
        filename = askopenfilename(initialdir=os.getcwd(),
                                   title='Select a file',
                                   defaultextension='.csv',
                                   filetypes=[('CSV', '.csv'),
                                              ('All Files', '*.*')])
        self.filenames[r] = filename
        width = 63
        filename = '...' + \
            filename[-width +
                     3:] if filename and len(filename) > width else filename
        self.label_list[r].config(text=filename)

    def delete_file(self, r):
        """ Deletes the specified file from storage and display.
        :param r: int type of data variable number.
        """
        self.filenames[r] = ''
        self.label_list[r].config(text='')

    def start_dataloader(self):
        """ Reads all the selected files, loads the data and passes it to 
        dataLoader.
        """

        # s = XML_handler()
        # s.new_config_file(self.save_path.name)
        # s.filename = self.save_path.name

        # self.s = XML_handler()
        # self.s.new_config_file()

        # self.s._print_xml_config()
        # self.s.load_XML(self.controller.output["xml_filename"])

        data = {}
        isVar = [0] * len(self.var)
        if len(self.label_list[0].cget("text")) > 0:
            for i, filename in enumerate(self.filenames):
                variable = self.var[i]
                if filename is not None and len(filename) > 0:
                    if filename.lower().endswith(('.csv')):
                        # Infers by default, should it be None?
                        data[variable] = pd.read_csv(filename)
                        isVar[i] = 1
                        self.controller.s.append_input_data(variable, rel_to_abs(filename))
                        if i == 0:
                            self.controller.Data.set(True)
                        if any(isVar[1::2]) and (
                                len(pd.unique(data[variable].to_numpy().flatten())) < len(data[variable].to_numpy().flatten())*0.2):
                            self.controller.output_type = 'classification' * \
                                all([float(i).is_integer()
                                    for i in data[variable].to_numpy().flatten()])
            if not any(isVar[1::2]):
                self.controller.output_type = 'unsupervised'
            self.newWindow.destroy()
            dataLoader(self.controller, data)
        else:
            tk.messagebox.showwarning(title='Error - X not specified',
                                      message='You need to specify X before proceeding.')

    def upload_data_path(self):
        """ Stores the directory containing the data that will be later loaded 
        """
        folder = askdirectory(initialdir=os.getcwd(),
                              title='Select a folder',
                              mustexist=True)
        if folder is not None and len(folder) > 0:
            self.controller.s.append_input_data('X', rel_to_abs(folder))
        
    def upload_data_folder(self):
        """ Stores the directory containing the data that will be later loaded 
        """
        folder = askdirectory(initialdir=os.getcwd(),
                              title='Select a folder',
                              mustexist=True)
        if folder is not None and len(folder) > 0:
            onlyfiles = [f for f in os.listdir(
                folder) if os.path.isfile(os.path.join(folder, f))]
            self.upload_data_file()
            for file in onlyfiles:
                name = file.lower()
                filename = os.path.join(folder, file)
                width = 63
                filenm = '...' + \
                    filename[-width +
                             3:] if filename and len(filename) > width else filename
                if name.endswith(('.csv')):
                    name = ''.join(ch for ch in name if ch.isalnum())
                    if 'test' in name or 'tst' in name:
                        if name[0] == 'x':
                            self.filenames[2] = filename
                            self.label_list[2].config(text=filenm)
                        elif name[0] == 'y':
                            self.filenames[3] = filename
                            self.label_list[3].config(text=filenm)
                    else:
                        if name[0] == 'x':
                            self.filenames[0] = filename
                            self.label_list[0].config(text=filenm)
                        elif name[0] == 'y':
                            self.filenames[1] = filename
                            self.label_list[1].config(text=filenm)



import os
import tkinter as tk
from typing import Dict
from tkinter import ttk

import numpy as np
import pandas as pd
from PIL import Image, ImageTk
from vai_lab._plugin_helpers import PluginSpecs
from vai_lab.Data.xml_handler import XML_handler
from vai_lab._import_helper import get_lib_parent_dir

_PLUGIN_READABLE_NAMES = {"progress_tracker":"default",
                            "progressTracker":"alias",
                            "progress tracker":"alias"}     # type:ignore
_PLUGIN_MODULE_OPTIONS = {"layer_priority": 2,              # type:ignore
                          "required_children": None}        # type:ignore
_PLUGIN_REQUIRED_SETTINGS = {"id_done"}                     # type:ignore
_PLUGIN_OPTIONAL_SETTINGS = {}                              # type:ignore


class progressTracker(tk.Frame):
    """Canvas for the visualisation of the pipeline state."""

    def __init__(self, parent, controller, config: dict):
        """ Here we define the frame displayed for the plugin specification."""

        super().__init__(parent, bg=parent['bg'])
        self.bg = parent['bg']
        self.controller = controller
        self.s = XML_handler()
        self.controller.title('Progress Tracker')
        
        script_dir = os.path.dirname(__file__)
        self.tk.call('wm', 'iconphoto', self.controller._w, ImageTk.PhotoImage(
            file=os.path.join(os.path.join(
                get_lib_parent_dir(),
                'utils',
                'resources',
                'Assets',
                'VAILabsIcon.ico'))))
        self.grid_columnconfigure(0, weight=1)

        frame1 = tk.Frame(self, bg=self.bg)
        self.frame2 = tk.Frame(self, bg=self.bg)
        frame3 = tk.Frame(self, bg=self.bg)
        self.frame4 = tk.Frame(self, bg=self.bg)

        # Create canvas
        self.width, self.height = 700, 700
        self.canvas = tk.Canvas(frame1, width=self.width,
                                height=self.height, background="white")
        self.canvas.pack(fill=tk.BOTH, expand=True, padx=(10, 0), pady=10)

        self.m: int
        self.w, self.h = 100, 50
        self.cr = 4
        """
        TODO: Implement clicking on canvas actions
        Commented out WIP code as it was complicating searching for bugs in plugincanvas
        """
        self.canvas.bind('<Button-1>', self.on_click)
        self.dataType: Dict = {}
        
        self.upload()

        sec = 5
        self.click = False
        self.my_label = tk.Label(self.frame2, 
                    text = 
                    'This window will get\nclosed after '+str(sec)+' seconds\nunless you click on the canvas.',
                    pady= 10,
                    font = self.controller.title_font,
                    bg = self.bg,
                    fg = 'white',
                    anchor = tk.CENTER)
        self.my_label.grid(column = 0,
                            row = 0, columnspan = 2, padx=10)
        tk.Button(
            self.frame4, text = 'Next step', fg = 'white', bg = parent['bg'], 
            height = 3, width = 15, font = self.controller.pages_font, 
            command = self.check_quit).grid(column = 1, row = 26, sticky="news", pady=(0,10))
        tk.Button(
            self.frame4, text = 'Stop pipeline', fg = 'white', bg = parent['bg'], 
            height = 3, width = 15, font = self.controller.pages_font, 
            command = self.terminate).grid(column = 0, row = 26, sticky="news", pady=(0,10))
        
        self.controller._append_to_output('close', False)
        self.save_path = ''
        self.saved = True
        frame1.grid(column=0, row=0, sticky="nsew")
        self.frame2.grid(column=1, row=0, sticky="new")
        frame3.grid(column=0, row=1, sticky="swe")
        self.frame4.grid(column=1, row=1, sticky="sew")
        
        frame3.grid_columnconfigure(tuple(range(2)), weight=1)
        self.frame4.grid_columnconfigure(tuple(range(2)), weight=1)

        self.controller.after(sec*1000,lambda:self.check_click())

    def class_list(self,value):
        """ Temporary fix """
        return value

    def on_click(self, event):
        """ Passes the mouse click coordinates to the select function."""
        self.click = True
        self.select(event.x, event.y)
    
    def check_click(self):
        """ If not clicked on canvas, closes the window."""
        if not self.click:
            self.check_quit()

    def terminate(self):
        """ Terminates window and pipeline. """
        self.controller._append_to_output('close', True)
        self.check_quit()
        
    def select(self, x: float, y:float):
        """ 
        Selects the module at the mouse location and updates the associated 
        plugins as well as the colours. 
        Blue means no plugin has been specified,
        Orange means the module is selected.
        Green means the plugin for this module is already set. 
        :param x: float type of module x coordinate
        :param y: float type of module y coordinate
        """
        self.selected = self.canvas.find_overlapping(
            x-5, y-5, x+5, y+5)
        if self.selected:
            if len(self.selected) > 2:
                self.canvas_selected = self.selected[-2]
            else:
                self.canvas_selected = self.selected[-1]
            if len(self.canvas.gettags(self.canvas_selected)) > 0:
                if not (len(self.canvas.gettags(self.canvas_selected)[0].split('-')) > 1) and\
                    not (self.canvas.gettags(self.canvas_selected)[0].split('-')[0] == 'loop'):
                    self.m = int(self.canvas.gettags(self.canvas_selected)[0][1:])
                if self.m > 1:
                    self.optionsWindow()

    def module_out(self, name):
        """ Updates the output DataFrame.

        :param name: name of the model
        :type name: str
        """
        name_list = list(self.out_data.columns)
        m_num = [n for n in name_list if (n.split('-')[0] == name)]
        if len(m_num) > 0:
            name_list.append(name + '-' + str(len(m_num)))
        else:
            name_list.append(name)
        self.canvas.itemconfig('t'+str(self.modules), text=name_list[-1])
        values = self.out_data.values
        values = np.vstack((
            np.hstack((values, np.zeros((values.shape[0], 1)))),
            np.zeros((1, values.shape[0]+1))))
        self.out_data = pd.DataFrame(values,
                                     columns=name_list,
                                     index=name_list)
        self.module_names.append(name_list[-1])

    def pretty_status(self,status: dict):
        status_str = ''
        for key in status.keys():
            status_str += str(key) +': '+str(status[key])+'\n'
        return status_str

    def add_module(self,boxName: str,x: float,y:float,ini = False,out = False):
        """ Creates a rectangular module with the corresponding text inside.

        :param boxName: name of the model
        :type boxName: str
        :param x: float type of module x coordinate
        :param y: float type of module y coordinate
        :param ini: bool type of whether the module corresponds to initialiser.
        :param out: bool type of whether the module corresponds to output.
        """
        if not ini and not out:
            tag = ('o'+str(self.modules),)
        else:  # Make initialisation and output unmoveable
            tag = ('n0',)
        text_w = self.controller.pages_font.measure(boxName+'-00') + 10
        # Check module status
        if 'start' in self.controller.status[boxName]:
            if 'finish' in self.controller.status[boxName]:
                colour = '#46da63'
            else:
                colour = '#dbaa21'
        else:
            colour = self.bg
        
        self.canvas.create_rectangle(
            x - text_w/2 , 
            y - self.h/2, 
            x + text_w/2, 
            y + self.h/2, 
            tags = tag + ('p'+str(self.modules),), 
            fill = colour, 
            width = 3,
            # activefill = '#dbaa21'
            )
        
        self.canvas.create_text(
            x, 
            y, 
            font = self.controller.pages_font, 
            text = boxName, 
            tags = tag + ('t'+str(self.modules),), 
            fill = '#d0d4d9', 
            justify = tk.CENTER)
        
        CanvasTooltip(self.canvas, self.canvas.find_withtag('p'+str(self.modules))[0], 
                           text = 'Model progress status:\n'+self.pretty_status(self.controller.status[boxName])) # Link box
        CanvasTooltip(self.canvas, self.canvas.find_withtag('t'+str(self.modules))[0], 
                           text = 'Model progress status:\n'+self.pretty_status(self.controller.status[boxName])) # Link text
        
        if not out:
            self.canvas.create_oval(
                x - self.cr,
                y + self.h/2 - self.cr,
                x + self.cr,
                y + self.h/2 + self.cr,
                width=2,
                fill='black',
                tags=tag + ('d'+str(self.modules),))

        if not ini:
            self.canvas.create_oval(
                x - self.cr,
                y - self.h/2 - self.cr,
                x + self.cr,
                y - self.h/2 + self.cr,
                width=2,
                fill='black',
                tags=tag + ('u'+str(self.modules),))

        if not out and not ini:
            self.canvas.create_oval(
                x - text_w/2 - self.cr,
                y - self.cr,
                x - text_w/2 + self.cr,
                y + self.cr,
                width=2,
                fill='black',
                tags=tag + ('l'+str(self.modules),))

            self.canvas.create_oval(
                x + text_w/2 - self.cr,
                y - self.cr,
                x + text_w/2 + self.cr,
                y + self.cr,
                width=2,
                fill='black',
                tags=tag + ('r'+str(self.modules),))

        self.canvas_startxy.append((x, y))
        self.connections[self.modules] = {}
        self.module_out(boxName)
        self.module_list.append(boxName)
        self.modules += 1
        
    def optionsWindow(self):
        """ Function to create a new window displaying the available options 
        of the selected plugin."""
        
        self.module = np.array(self.module_list)[self.m == np.array(self.id_mod)][0]
        self.plugin = self.s.loaded_modules[self.module]['plugin']['plugin_name']
        module_type = self.s.loaded_modules[self.module]['module_type']
        ps = PluginSpecs()
        self.opt_settings = ps.optional_settings[module_type][self.plugin]
        self.req_settings = ps.required_settings[module_type][self.plugin]
        if (len(self.opt_settings) != 0) or (len(self.req_settings) != 0):
            if hasattr(self, 'newWindow') and (self.newWindow!= None):
                self.newWindow.destroy()
            self.newWindow = tk.Toplevel(self.controller)
            # Window options
            self.newWindow.title(self.plugin+' plugin options')
            script_dir = get_lib_parent_dir()
            self.tk.call('wm', 'iconphoto', self.newWindow, ImageTk.PhotoImage(
                file=os.path.join(os.path.join(
                    script_dir,
                    'utils',
                    'resources',
                    'Assets',
                    'VAILabsIcon.ico'))))
            self.newWindow.geometry("350x400")

            frame1 = tk.Frame(self.newWindow)
            frame4 = tk.Frame(self.newWindow)
            
            # Print settings
            tk.Label(frame1,
                  text ="Please indicate your desired options for the "+self.plugin+" plugin.", anchor = tk.N, justify=tk.LEFT).pack(expand = True)

            style = ttk.Style()
            style.configure(
                "Treeview", background='white', foreground='white',
                rowheight=25, fieldbackground='white',
                # font=self.controller.pages_font
                )
            style.configure("Treeview.Heading", 
            # font=self.controller.pages_font
            )
            style.map('Treeview', background=[('selected', 'grey')])

            frame2 = tk.Frame(self.newWindow, bg='green')
            self.r = 1
            self.create_treeView(frame2)
            self.fill_treeview(frame2, self.req_settings, self.opt_settings)
            frame2.grid(column=0, row=1, sticky="nswe", pady=10, padx=10)

            frame2.grid_rowconfigure(tuple(range(self.r)), weight=1)
            frame2.grid_columnconfigure(tuple(range(2)), weight=1)

            self.finishButton = tk.Button(
                frame4, text='Finish', command=self.removewindow)
            self.finishButton.grid(
                column=1, row=0, sticky="es", pady=(0, 10), padx=(0, 10))
            self.finishButton.bind(
                "<Return>", lambda event: self.removewindow())
            self.newWindow.protocol('WM_DELETE_WINDOW', self.removewindow)

            frame1.grid(column=0, row=0, sticky="ew")
            frame4.grid(column=0, row=2, sticky="se")
            self.newWindow.grid_rowconfigure(1, weight=2)
            self.newWindow.grid_columnconfigure(0, weight=1)

    def create_treeView(self, tree_frame):
        """ Function to create a new tree view in the given frame

        Parameters
        ----------
        tree_frame : tk.Frame
                    frame where the tree view will be included
        key : str
              key for the tree dictionary
        """

        tree_scrollx = tk.Scrollbar(tree_frame, orient='horizontal')
        tree_scrollx.pack(side=tk.BOTTOM, fill=tk.X)
        tree_scrolly = tk.Scrollbar(tree_frame)
        tree_scrolly.pack(side=tk.RIGHT, fill=tk.Y)

        self.tree = ttk.Treeview(tree_frame,
                                yscrollcommand=tree_scrolly.set,
                                xscrollcommand=tree_scrollx.set)
        self.tree.pack(fill='both', expand=True)

        tree_scrollx.config(command=self.tree.xview)
        tree_scrolly.config(command=self.tree.yview)

        columns_names = ['Name', 'Type', 'Value']
        self.tree['columns'] = columns_names

        # Format columns
        self.tree.column("#0", width=20,
                        minwidth=0, stretch=tk.NO)
        for n, cl in enumerate(columns_names):
            self.tree.column(
                cl, width=int(self.controller.pages_font.measure(str(cl)))+20,
                minwidth=50, anchor=tk.CENTER)
        # Headings
        for cl in columns_names:
            self.tree.heading(cl, text=cl, anchor=tk.CENTER)
        self.tree.tag_configure('req', foreground='black',
                                background='#9fc5e8')
        self.tree.tag_configure('opt', foreground='black',
                                background='#cfe2f3')
        self.tree.tag_configure('type', foreground='black',
                                background='#E8E8E8')
        self.tree.tag_configure('func', foreground='black',
                                background='#DFDFDF')
        # Define double-click on row action
        self.tree.bind("<Double-1>", self.OnDoubleClick)

    def OnDoubleClick(self, event):
        """ Executed when a row of the treeview is double clicked.
        Opens an entry box to edit a cell. """

        # ii = self.notebook.index(self.notebook.select())
        self.treerow = self.tree.identify_row(event.y)
        self.treecol = self.tree.identify_column(event.x)
        tags = self.tree.item(self.treerow)["tags"]
        if len(tags) > 0 and tags[0] in ['opt', 'req']:
            # get column position info
            x, y, width, height = self.tree.bbox(self.treerow, self.treecol)

            # y-axis offset
            pady = height // 2
            # pady = 0

            if hasattr(self, 'entry'):
                self.entry.destroy()

            self.entry = tk.Entry(self.tree, justify='center')

            if int(self.treecol[1:]) > 0:
                value = self.tree.item(self.treerow)['values'][int(str(self.treecol[1:]))-1] 
                value = str(value) if str(value) not in ['default', 'Choose X or Y'] else ''
                self.entry.insert(0, value)
                # self.entry['selectbackground'] = '#123456'
                self.entry['exportselection'] = False

                self.entry.focus_force()
                self.entry.bind("<Return>", self.on_return)
                self.entry.bind("<Escape>", lambda *ignore: self.entry.destroy())

                self.entry.place(x=x,
                                y=y + pady,
                                anchor=tk.W, width=width)

    def on_return(self, event):
        """ Executed when the entry is edited and pressed enter.
        Saves the edited value"""

        val = self.tree.item(self.treerow)['values']
        val[int(self.treecol[1:])-1] = self.entry.get()
        if self.entry.get() != '':
            self.tree.item(self.treerow, values=tuple([val[0], val[1], self.entry.get()]))
        elif val[2] == '':
            self.tree.item(self.treerow, values=tuple([val[0], val[1], 'default']))
        else:
            self.tree.item(self.treerow, values=val)
        self.entry.destroy()
        self.saved = False

    def fill_treeview(self, frame, req_settings, opt_settings, parent = ''):
        """ Adds an entry for each setting. Displays it in the specified row.
        :param req_settings: dict type of plugin required setting options
        :param opt_settings: dict type of plugin optional setting options
        :param parent: string type of parent name
        """
        self.tree.insert(parent=parent, index='end', iid=parent+'_req', text='',
            values=tuple(['Required settings', '', '']), tags=('type',))
        self.r+=1
        for arg, val in req_settings.items():
            if arg == 'Data':
                self.tree.insert(parent=parent+'_req', index='end', iid=str(self.r), text='',
                    values=tuple([arg, val, 'Choose X or Y']), tags=('req',))
            else:
                self.tree.insert(parent=parent+'_req', index='end', iid=str(self.r), text='',
                                    values=tuple([arg, val, '']), tags=('req',))
            self.r+=1
        self.tree.insert(parent=parent, index='end', iid=parent+'_opt', text='',
            values=tuple(['Optional settings', '', '']), tags=('type',))
        self.r+=1
        for arg, val in opt_settings.items():
            self.tree.insert(parent=parent+'_opt', index='end', iid=str(self.r), text='',
                                 values=tuple([arg, val, 'default']), tags=('opt',))
            self.r+=1

    def removewindow(self):
        """ Stores settings options and closes window """
        self.req_settings.pop("Data", None)
        children = self.get_all_children()
        for child in children:
            tag = self.tree.item(child)["tags"][0]            
            if tag in ['req', 'opt']:
                val = self.tree.item(child)["values"]
                self.settingOptions(tag, val) 
        self.newWindow.destroy()
        self.newWindow = None
        self.focus()
    
    def get_all_children(self, item=""):
        """ Iterates over the treeview to get all childer """
        children = self.tree.get_children(item)
        for child in children:
            children += self.get_all_children(child)
        return children

    def settingOptions(self, tag, val):
        """ Identifies how the data should be stored """
        if val[0] == 'Data':
            if val[2] == 'Choose X or Y' or len(val[2]) == 0:
                self.updateSettings(tag, val[0], 'X')
            else:
                self.updateSettings(tag, val[0], val[2])
        else:
            if val[2] == 'default' or len(str(val[2])) == 0:
                self.updateSettings(tag, val[0])
            else:
                self.updateSettings(tag, val[0], val[2])

    def updateSettings(self, tag, key, value = None):
        """ Return the selected settings 

        Parameters
        ----------
        tag : str
              tag for the settings
        """
        if tag == 'req':
            if value is not None or self.req_settings[key] != value:
                self.req_settings[key] = value
            else:
                self.req_settings.pop(key, None)
        elif tag == 'opt':
            if value is not None or self.opt_settings[key] != value:
                self.opt_settings[key] = value
            else:
                self.opt_settings.pop(key, None)

    def on_return_entry(self, r):
        """ Changes focus to the next available entry. When no more, focuses 
        on the finish button.
        : param r: int type of entry id.
        """
        if r < len(self.entry):
            self.entry[r].focus()
        else:
            self.finishButton.focus()

    def upload(self):
        """ Opens the XML file that was previously uploaded and places the 
        modules, loops and connections in the canvas."""

        filename = self.controller.output["xml_filename"]

        self.reset()

        self.s = XML_handler()
        self.s.load_XML(filename)
        # self.s._print_pretty(self.s.loaded_modules)
        modules = self.s.loaded_modules
        modout = modules['Output']
        del modules['Initialiser'], modules['Output'] # They are generated when resetting
        self.disp_mod = ['Initialiser', 'Output']
        self.id_mod = [0, 1]

        # Place the modules
        self.place_modules(modules)
        connect = list(modout['coordinates'][2].keys())
        for p, parent in enumerate(modout['parents']):
            parent_id = self.id_mod[np.where(
                np.array(self.disp_mod) == parent)[0][0]]
            out, ins = modout['coordinates'][2][connect[p]].split('-')
            xout, yout, _, _ = self.canvas.coords(out[0]+str(parent_id))
            xins, yins, _, _ = self.canvas.coords(ins[0]+str(1))
            self.canvas.create_line(
                xout + self.cr,
                yout + self.cr,
                xins + self.cr,
                yins + self.cr,
                fill="red",
                arrow=tk.LAST,
                tags=('o'+str(parent_id),
                      'o'+str(1), modout['coordinates'][2][connect[p]]))
            self.out_data.iloc[int(parent_id)][1] = 1
            self.connections[1][
                int(parent_id)] = out[0]+str(parent_id) + '-' + ins[0]+str(1)
        self.m = self.id_mod[2]
        x0, y0, x1, y1 = self.canvas.coords('p'+str(self.m))
        # self.select(x0, y0)

    def place_modules(self, modules: dict):
        """Places the modules in the dictionary in the canvas.
        :param modules: dict type of modules in the pipeline.
        """

        for key in [key for key, val in modules.items() if type(val) == dict]:
            if modules[key]['class'] == 'loop':
                # Extracts numbers from string
                l = int(
                    ''.join(map(str, list(filter(str.isdigit, modules[key]['name'])))))
                x0, y0, x1, y1 = modules[key]['coordinates']
                self.canvas.create_rectangle(x0, y0,
                                             x1, y1,
                                             outline='#4ff07a',
                                             tag='loop-'+str(l)) # type:ignore
                text_w = self.controller.pages_font.measure(
                    modules[key]['type']) + 20
                self.canvas.create_text(
                    x0 + text_w/2,
                    y0 + 20,
                    font=self.controller.pages_font,
                    text=modules[key]['type'],
                    tags=('loop-'+str(l), 'type'+'-'+str(l)),
                    justify=tk.CENTER)
                text_w = self.controller.pages_font.measure(
                    modules[key]['condition']) + 20
                self.canvas.create_text(
                    x1 - text_w/2,
                    y0 + 20,
                    font=self.controller.pages_font,
                    text=modules[key]['condition'],
                    tags=('loop-'+str(l), 'condition'+'-'+str(l)),
                    justify=tk.CENTER)
                self.loops.append({'type': modules[key]['type'],
                                   'condition': modules[key]['condition'],
                                   'mod': [],
                                   'coord': (x0, y0,
                                             x1, y1)})
                self.place_modules(modules[key])
            else:
                # Display module
                self.add_module(key,
                                modules[key]['coordinates'][0][0],
                                modules[key]['coordinates'][0][1])
                # self.module_list[-1] = modules[key]['module_type']
                self.id_mod.append(modules[key]['coordinates'][1])
                connect = list(modules[key]['coordinates'][2].keys())

                # Connect modules
                for p, parent in enumerate(modules[key]['parents']):
                    if not (parent[:4] == 'loop'):
                        parent_id = self.id_mod[np.where(
                            np.array(self.disp_mod) == parent)[0][0]]
                        out, ins = modules[key]['coordinates'][2][connect[p]].split(
                            '-')
                        xout, yout, _, _ = self.canvas.coords(
                            out[0]+str(parent_id))
                        xins, yins, _, _ = self.canvas.coords(
                            ins[0]+str(self.id_mod[-1]))
                        self.canvas.create_line(
                            xout + self.cr,
                            yout + self.cr,
                            xins + self.cr,
                            yins + self.cr,
                            fill="red",
                            arrow=tk.LAST,
                            tags=('o'+str(parent_id),
                                  'o'+str(self.id_mod[-1]), modules[key]['coordinates'][2][connect[p]]))
                        self.out_data.iloc[int(parent_id)][int(
                            self.id_mod[-1])] = 1
                        self.connections[int(self.id_mod[-1])][
                            int(parent_id)] = out[0]+str(parent_id) + '-' + ins[0]+str(self.id_mod[-1])
                    else:
                        self.loops[-1]['mod'].append(key)
                self.disp_mod.append(key)

    def reset(self):
        """ Resets the canvas and the stored information."""
        self.canvas.delete(tk.ALL)  # Reset canvas

        if hasattr(self, 'newWindow') and (self.newWindow != None):
            self.newWindow.destroy()
        
        self.canvas_startxy = []
        self.out_data = pd.DataFrame()
        self.connections = {}
        self.modules = 0
        self.module_list = []
        self.module_names = []

        self.add_module('Initialiser', self.width/2, self.h, ini=True)
        self.add_module('Output', self.width/2, self.height - self.h, out=True)

        self.draw = False
        self.loops = []
        self.drawLoop = False
        self.l = 0
        self.id_done = [0, 1]
        self.plugin = {}

    def check_quit(self):
        self.controller.destroy()
        
class CanvasTooltip:
    '''
    It creates a tooltip for a given canvas tag or id as the mouse is
    above it.

    This class has been derived from the original Tooltip class I updated
    and posted back to StackOverflow at the following link:

    https://stackoverflow.com/questions/3221956/
           what-is-the-simplest-way-to-make-tooltips-in-tkinter/
           41079350#41079350

    Alberto Vassena on 2016.12.10.
    '''

    def __init__(self, canvas, tag_or_id,
                 *,
                 bg='#FFFFEA',
                 pad=(5, 3, 5, 3),
                 text='canvas info',
                 waittime=100,
                 wraplength=250):
        self.waittime = waittime  # in miliseconds, originally 500
        self.wraplength = wraplength  # in pixels, originally 180
        self.canvas = canvas
        self.text = text
        self.canvas.tag_bind(tag_or_id, "<Enter>", self.onEnter)
        self.canvas.tag_bind(tag_or_id, "<Leave>", self.onLeave)
        self.canvas.tag_bind(tag_or_id, "<ButtonPress>", self.onLeave)
        self.bg = bg
        self.pad = pad
        self.id = None
        self.tw = None

    def onEnter(self, event=None):
        self.schedule()

    def onLeave(self, event=None):
        self.unschedule()
        self.hide()

    def schedule(self):
        self.unschedule()
        self.id = self.canvas.after(self.waittime, self.show)

    def unschedule(self):
        id_ = self.id
        self.id = None
        if id_:
            self.canvas.after_cancel(id_)

    def show(self, event=None):
        def tip_pos_calculator(canvas, label,
                               *,
                               tip_delta=(10, 5), pad=(5, 3, 5, 3)):

            c = canvas

            s_width, s_height = c.winfo_screenwidth(), c.winfo_screenheight()

            width, height = (pad[0] + label.winfo_reqwidth() + pad[2],
                             pad[1] + label.winfo_reqheight() + pad[3])

            mouse_x, mouse_y = c.winfo_pointerxy()

            x1, y1 = mouse_x + tip_delta[0], mouse_y + tip_delta[1]
            x2, y2 = x1 + width, y1 + height

            x_delta = x2 - s_width
            if x_delta < 0:
                x_delta = 0
            y_delta = y2 - s_height
            if y_delta < 0:
                y_delta = 0

            offscreen = (x_delta, y_delta) != (0, 0)

            if offscreen:

                if x_delta:
                    x1 = mouse_x - tip_delta[0] - width

                if y_delta:
                    y1 = mouse_y - tip_delta[1] - height

            offscreen_again = y1 < 0  # out on the top

            if offscreen_again:
                # No further checks will be done.

                # TIP:
                # A further mod might automagically augment the
                # wraplength when the tooltip is too high to be
                # kept inside the screen.
                y1 = 0

            return x1, y1

        bg = self.bg
        pad = self.pad
        canvas = self.canvas

        # creates a toplevel window
        self.tw = tk.Toplevel(canvas.master)

        # Leaves only the label and removes the app window
        self.tw.wm_overrideredirect(True)

        win = tk.Frame(self.tw,
                       background=bg,
                       borderwidth=0)
        label = tk.Label(win,
                          text=self.text,
                          justify=tk.LEFT,
                          background=bg,
                          relief=tk.SOLID,
                          borderwidth=0,
                          wraplength=self.wraplength)

        label.grid(padx=(pad[0], pad[2]),
                   pady=(pad[1], pad[3]),
                   sticky=tk.NSEW)
        win.grid()

        x, y = tip_pos_calculator(canvas, label)

        self.tw.wm_geometry("+%d+%d" % (x, y))

    def hide(self):
        if self.tw:
            self.tw.destroy()
        self.tw = None

if __name__ == "__main__":
    app = progressTracker()
    app.mainloop()



from attr import attr
from vai_lab._plugin_templates import EnvironmentPluginT
from vai_lab._import_helper import rel_to_abs
from typing import Any, Dict
import pybullet as p
from time import sleep


_PLUGIN_READABLE_NAMES = {"PyBullet":"default"}                                                 # type:ignore
_PLUGIN_MODULE_OPTIONS = {}                                                                     # type:ignore
_PLUGIN_REQUIRED_SETTINGS = {"model_dir": "str", "headless": "bool",
                                "timestep":"float", "max_steps":"int"}                          # type:ignore
_PLUGIN_OPTIONAL_SETTINGS = {"gravity":"list"}                                                  # type:ignore
_PLUGIN_REQUIRED_DATA = {}                                                                      # type:ignore
_PLUGIN_OPTIONAL_DATA = {"X","Y","X_tst", 'Y_tst'}                                              # type:ignore


class PyBulletEnv(EnvironmentPluginT):
    """
    Loads the pybullet library as wildcard and exposes all functions
    """

    def __init__(self) -> None:
        super().__init__(globals())
        self.connection_mode = p.GUI
        self.model_ids: Dict = {}

    def set_gui(self, use_gui: bool = True):
        if use_gui:
            self.connection_mode = p.GUI  # Use pybullet GUI
        else:
            self.connection_mode = p.DIRECT  # Use pybullet without GUI

    def connect(self):
        self.physicsClient = p.connect(self.connection_mode)

    def disconnect(self):
        return p.disconnect()

    def reset(self):
        return p.resetSimulation()

    def load_pb_data(self):
        import pybullet_data                                        # type:ignore
        p.setAdditionalSearchPath(pybullet_data.getDataPath())

    def _load_model_by_type(self, model):
        model = rel_to_abs(model)
        ext = model.split(".")[-1]
        name = model.split("/")[-1].split(".")[0]
        if name == "plane":
            self.load_pb_data()
        if ext == "urdf":
            self.model_ids[name] = p.loadURDF(model)
        elif ext == "sdf":
            self.model_ids[name] = p.loadSDF(model, 1, 1.0)
        elif ext == "xml":
            self.model_ids[name] = p.loadMJCF(model)

    def load_model(self) -> None:
        model_paths = self._config["options"]["model_dir"]
        if type(model_paths) == str:
            self._load_model_by_type(model_paths)
        elif type(model_paths) == list:
            for model in model_paths:
                self._load_model_by_type(model)

    def _set_options(self):
        api_list = dir(p)
        for key, value in self._config["options"].items():
            if key in api_list:
                getattr(p,key)(*value)
        if "timestep" in self._config["options"]:
            p.setPhysicsEngineParameter(
                fixedTimeStep=self._config["options"]["timestep"])
        elif "max_steps" in self._config["options"]:
            p.setPhysicsEngineParameter(
                numSolverIterations=self._config["options"]["max_steps"])

    def run_simulation(self):
        self._set_options()
        for step in range(1, self._config["options"]["max_steps"]):
            p.stepSimulation()
            sleep(self._config["options"]["timestep"])
        self.disconnect()

    def __getattr__(self, attr: str) -> Any:
        """Allows calling pybullet functions directly as if they were functions of this class

        TODO: This is probably not the best way to do this, but pybullet is
        a compiled module and cannot be used as a parent. This is a workaround.
        """
        return getattr(p, attr)


if __name__ == "__main__":
    pb = PyBulletEnv()
    print(pb.ACTIVATION_STATE_DISABLE_SLEEPING)
    # pb.set_gui(True)
    # pb.connect()



from os import path
from ast import literal_eval
from typing import Any, Dict, List, Optional, Union
import xml.etree.ElementTree as ET

# Allows standalone testing 
if not __package__:
    import sys
    root_mod = path.dirname(path.dirname(path.dirname(__file__)))
    sys.path.append(root_mod)

from vai_lab._import_helper import get_lib_parent_dir


class XML_handler:
    def __init__(self, filename: str = None):
        """Loads XML file and parses into nested dictionaries
        This module is used as the backbone for all future modules.
        If filename is not given, can be passed straight to self.load_XML()
        :param filename [optional]: filename from which to load XML
        """
        self.lib_base_path = get_lib_parent_dir()
        self.loaded_modules: Dict[str, Dict] = {}

        """valid_tags lists the available XML tags and their function
        TODO: populate the modules in this list automatically
        TODO: Can extend the plugin_helpers.py script to automatically populate these things
        """
        self._valid_tags: Dict[str, str] = {
            "pipeline": "declaration",
            "relationships": "relationships",
            "plugin": "plugin",
            "coordinates": "list",
            "Initialiser": "entry_point",
            "inputdata": "data",
            "Output": "exit_point",
            "UserInteraction": "module",
            "DataProcessing": "module",
            "Modelling": "module",
            "InputData": "module",
            "DecisionMaking": "module",
            "Environment": "module",
            "DataStorage": "module",
            "loop": "loop"
        }

        if filename is not None:
            self.load_XML(filename)

    def _check_filename_abs_rel(self, filename: str):
        """Checks if path is relative or absolute
        If absolute, returns original path 
        If relative, converts path to absolute by appending to base directory
        """
        return path.normpath(path.join(self.lib_base_path, filename))

    def _check_file_exists(self, filename: str):
        """Checks if a given path exists
        Convert any rel path to abs first
        :returns: 
            - True if file exists
            - False if file does not exist
        """
        abs_path = self._check_filename_abs_rel(filename)
        return path.exists(abs_path)

    def set_filename(self, filename: str):
        """Sets filename. Converts relative paths to absolute first."""
        self.filename: str = self._check_filename_abs_rel(filename)

    def append_initialiser(self):
        self.append_pipeline_module("Initialiser",
                                    "Initialiser",
                                    None,
                                    None,
                                    [],
                                    [],
                                    None,
                                    [0, 0, 0])

    def new_config_file(self, filename: str = None) -> None:
        """Constructs new XML file with minimal format"""
        if filename is not None:
            self.set_filename(filename)
        self.tree = ET.ElementTree(ET.Element("pipeline"))
        self.root = self.tree.getroot()
        self.append_initialiser()

    def load_XML(self, filename: str) -> None:
        """Loads XML file into class. 
        """
        if filename != None:
            self.set_filename(filename)
        self.tree = ET.parse(self.filename)
        self._parse_XML()

    def _parse_XML(self) -> None:
        self.root = self.tree.getroot()
        self._parse_tags(self.root, self.loaded_modules)

    def _bullet_list(self, val_list: Union[List[Any], Dict[str, str]]) -> str:
        sep = "\n\t\u2022"
        return sep + " {}".format(("," + sep + " ").join([*val_list]))

    def _parse_tags(self, element: ET.Element, parent: dict) -> None:
        """Detect tags and send them to correct method for parsing
        Uses getattr to call the correct method
        :param elem: xml.etree.ElementTree.Element to be parsed
        :param parent: dict or dict fragment parsed tags will be appened to
        """
        for child in element:
            try:
                tag_type = self._valid_tags[child.tag]
                getattr(self, "_load_{}".format(tag_type))(child, parent)
            except KeyError:
                from sys import exit
                print("\nError: Invalid XML Tag." +
                      "\nXML tag \"{0}\"".format(child.tag) +
                      "in \"{0}\" not found".format(element.tag) +
                      "\nValid tags are:" +
                      self._bullet_list(self._valid_tags))
                exit(1)
            except AssertionError as Error:
                from sys import exit
                print(Error)
                exit(1)

    def _load_module(self, element: ET.Element, parent: dict) -> None:
        """Parses tags associated with modules and appends to parent dict
        :param elem: xml.etree.ElementTree.Element to be parsed
        :param parent: dict or dict fragment parsed tags will be appened to
        """
        module_name = element.attrib["name"]
        module_type = element.tag
        parent[module_name] = {"name": module_name,
                               "class": self._valid_tags[module_type],
                               "module_type": module_type}
        self._parse_tags(element, parent[module_name])

    def _load_plugin(self, element: ET.Element, parent: dict) -> None:
        """Parses tags associated with plugins and appends to parent dict
        :param elem: xml.etree.ElementTree.Element to be parsed
        :param parent: dict or dict fragment parsed tags will be appened to
        """
        parent["plugin"] = {}
        parent["plugin"]["plugin_name"] = element.attrib["type"]
        parent["plugin"]["options"] = {}
        for child in element:
            if child.text is not None:
                val = self._parse_text_to_list(child)
                val = (val[0] if len(val) == 1 else val)
                parent["plugin"]["options"][child.tag] = val
            for key in child.attrib:
                if key == "val":
                    parent["plugin"]["options"][child.tag] = child.attrib[key]
                else:
                    parent["plugin"]["options"][child.tag] = {
                        key: child.attrib[key]}

    def _load_entry_point(self, element: ET.Element, parent: dict) -> None:
        """Parses tags associated with initialiser and appends to parent dict
        :param elem: xml.etree.ElementTree.Element to be parsed
        :param parent: dict or dict fragment parsed tags will be appened to
        """
        initialiser_name = element.attrib["name"]
        parent[initialiser_name] = {"name": initialiser_name,
                                    "class": self._valid_tags[element.tag]}
        self._parse_tags(element, parent[initialiser_name])

    def _load_data(self, element: ET.Element, parent: dict) -> None:
        """Parses tags associated with initial data files and appends to parent dict
        :param elem: xml.etree.ElementTree.Element to be parsed
        :param parent: dict or dict fragment parsed tags will be appened to
        """
        data_name = (element.attrib["name"]
                     if "name" in element.attrib else "input_data")
        parent[data_name] = {"name": data_name,
                             "class": "data",
                             "to_load": {}}
        for child in element:
            assert "file" in child.attrib or "folder" in child.attrib,\
                str("XML Parse Error \
                        \n\tA path to data file must be specified. \
                        \n\t{0} does not contain the \"file\" tag. \
                        \n\tCorrect usage: {0} file = <path-to-file>".format(child.tag))

            if "file" in child.attrib:
                assert self._check_file_exists(child.attrib["file"]),\
                    str("Error: Data file not found. \
                        \n\tFile: {0} does not exist"
                        .format(self._check_filename_abs_rel(child.attrib["file"])))
                parent[data_name]["to_load"][child.tag] = child.attrib["file"]

            if "folder" in child.attrib:
                assert self._check_file_exists(child.attrib["folder"]),\
                    str("Error: Data folder not found. \
                        \n\tFolder: {0} does not exist"
                        .format(self._check_filename_abs_rel(child.attrib["folder"])))
                parent[data_name]["to_load"][child.tag] = child.attrib["folder"]

    def _load_exit_point(self, element: ET.Element, parent: dict) -> None:
        """Parses tags associated with output and appends to parent dict
        :param elem: xml.etree.ElementTree.Element to be parsed
        :param parent: dict or dict fragment parsed tags will be appened to
        """
        parent["Output"] = {"name": "Output",
                            "class": self._valid_tags[element.tag]}
        self._parse_tags(element, parent["Output"])

    def _load_loop(self, element: ET.Element, parent: dict) -> None:
        """Parses tags associated with loops and appends to parent dict
        :param elem: xml.etree.ElementTree.Element to be parsed
        :param parent: dict or dict fragment parsed tags will be appened to
        """
        loop_name = element.attrib["name"]
        parent[loop_name] = {
            "name": loop_name,
            "class": self._valid_tags[element.tag],
            "type": element.attrib["type"].lower(),
            "condition": element.attrib["condition"],
        }
        self._parse_tags(element, parent[loop_name])

    def _load_relationships(self, element: ET.Element, parent: dict) -> None:
        """Parses tags associated with relationships and adds to parent dict
        :param elem: xml.etree.ElementTree.Element to be parsed
        :param parent: dict or dict fragment parsed tags will be appened to
        """
        parent["parents"] = []
        parent["children"] = []
        for rel in element:
            if rel.tag == "parent":
                parent["parents"].append(rel.attrib["name"])
            elif rel.tag == "child":
                parent["children"].append(rel.attrib["name"])

    def _load_list(self, element: ET.Element, parent: dict) -> None:
        """Parses elements consisting of lists, e.g. coordinates
        :param elem: xml.etree.ElementTree.Element to be parsed
        :param parent: dict or dict fragment parsed tags will be appened to
        """
        if element.text is not None:
            parent[element.tag] = self._parse_text_to_list(element)
            if len(parent[element.tag]) == 1:
                parent[element.tag] = parent[element.tag][0]

    def _str_to_num(self, string:str) -> Union[str, int, float]:
        """Converts a string containing only numeric data to float or int
            Otherwise returns string

        :param string: The string to be converted if possible
        :returns out: either a string, float or int
        """
        try:
            return int(string)
        except ValueError:
            pass

        try:
            return float(string)
        except ValueError:
            return string

    def _parse_text_to_list(self, element: ET.Element) -> List:
        """Formats raw text data
        :param elem: xml.etree.ElementTree.Element to be parsed
        :returns out: list containing parsed text data
        """
        if element.text is not None:
            new = element.text.strip().replace(" ", "")

        out: List[Any] = new.split("\n")
        raw_elem_text = str()
        for idx in range(0, len(out)):
            raw_elem_text = (raw_elem_text+"\n{}").format(out[idx])
            if "[" in out[idx] and "]" in out[idx]:
                out[idx] = literal_eval(out[idx])
            if "(" in out[idx] and ")" in out[idx]:
                out[idx] = list(literal_eval(out[idx]))
            if type(out[idx]) == str:
                out[idx] = self._str_to_num(out[idx])
        element.text = raw_elem_text
        return out

    def _print_pretty(self, element: ET.Element):
        """Print indented dictionary to screen"""
        from pprint import PrettyPrinter
        pp = PrettyPrinter(sort_dicts=False, width=100)
        pp.pprint(element)

    def _print_xml_config(self):
        """Print indented pipeline specifications to screen"""
        if len(self.loaded_modules) == 0:
            self._parse_XML()
        self._print_pretty(self.loaded_modules)

    def write_to_XML(self):
        """Formats XML Tree correctly, then writes to filename
        TODO: add overwrite check and alternate filename option
        """
        self._indent(self.root)
        self.tree.write(self.filename)

    def _indent(self, elem: ET.Element, level: int = 0):
        """Formats XML tree to be human-readable before writing to file
        :param elem: xml.etree.ElementTree.Element to be indented
        :param level: int, level of initial indentation
        """
        sep = "    "
        i = "\n" + level*sep
        if len(elem):
            if not elem.text or not elem.text.strip():
                elem.text = i + sep
            if not elem.tail or not elem.tail.strip():
                elem.tail = i
            for elem in elem:
                self._indent(elem, level+1)
            if not elem.tail or not elem.tail.strip():
                elem.tail = i
        else:
            if level and (not elem.tail or not elem.tail.strip()):
                elem.tail = i
            if elem.text:
                elem.text = elem.text.replace("\n", ("\n" + (level+1)*sep))
                elem.text = ("{0}{1}").format(elem.text, i)

    def _find_dict_with_key_val_pair(self,
                                     parent: dict,
                                     key: str,
                                     val: Any,
                                     out=None) -> List[Dict[str, str]]:
        """Seach nested dict for a given key-value pair
        :param parent: dict to be searched
        :param key: str of key to find
        :param val: val to find in target key
        :param out [optional]: list to append found dicts to
        :returns out: list of dicts containing val pairs if any
        """
        if out is None:
            out = []
        for k in parent.keys():
            if isinstance(parent[k], dict):
                if key in parent[k]:
                    if parent[k][key] == val:
                        out.append(parent[k])
            if isinstance(parent[k], dict):
                self._find_dict_with_key_val_pair(parent[k], key, val, out)
        return out

    def _get_element_from_name(self, name: Union[str, Optional[str]]) -> ET.Element:
        """Find a module name and return its parent tags"""
        if name == None:
            return self.root

        elems = self.root.findall(".//*[@name='{0}']".format(name))

        unique_elem = [e for e in elems if e.tag !=
                       "parent" and e.tag != "child"]

        assert len(unique_elem) < 2,\
            "Error: More than one tag with same identifier"

        assert len(unique_elem) > 0, \
            "Error: No element exists with name \"{0}\"".format(name)

        return unique_elem[0]

    def _get_all_elements_with_tag(self, tag: str):
        """Return all elements with a given tag
        :param tag: string with tag name
        """
        elems = self.root.findall(".//*{0}".format(tag))
        return elems

    def _loop_rels_autofill(self,
                            elem: ET.Element,
                            xml_child: str):
        """Add the name of a new nested module to the relationships of a loop
        :param elem: ET.Element
        :param xml_child:str name of child to be nested in XML
        """
        if elem.tag != "loop":
            return elem
        else:
            return self._add_relationships(elem, [], [xml_child])

    def _add_relationships(self,
                           elem: ET.Element,
                           parents: list,
                           children: list):
        """Add parents and/or children as a subelement
        Checks for duplicates before appending
        :param elem: ET.Element to which the relationship are appeneded
        :param parents: list of strings of parents to be appended
        :param children: list of strings of children to be appended
        """
        rels_elem = elem.find("./relationships")
        if rels_elem is None:
            rels_elem = ET.SubElement(elem, "relationships")
        for p in parents:
            if not elem.findall(".//parent/[@name='{0}']".format(p)):
                new_parent = ET.SubElement(rels_elem, "parent")
                new_parent.set('name', p)
        for c in children:
            if not elem.findall(".//child/[@name='{0}']".format(c)):
                new_child = ET.SubElement(rels_elem, "child")
                new_child.set('name', c)
        return elem

    def append_module_relationships(self,
                                    module_name: str,
                                    parents: list,
                                    children: list):
        elem = self._get_element_from_name(module_name)
        self._add_relationships(elem, parents, children)

    def update_module_coords(self,
                             module_name: str,
                             coords: list = None,
                             save_changes: bool = False):
        elem = self._get_element_from_name(module_name)
        self._add_coords(elem, coords)
        self._parse_XML()
        if save_changes:
            self.write_to_XML()

    def _add_coords(self,
                    elem: ET.Element,
                    coords: list = None):
        if coords == None:
            return elem
        coords_elem = elem.find("./coordinates")
        if coords_elem is None:
            coords_elem = ET.SubElement(elem, "coordinates")
        coords_elem.text = str("\n{0}".format(coords))
        return elem

    def update_plugin_options(self,
                              xml_parent_name: Union[ET.Element,str],
                              options: dict,
                              save_changes: bool = False):
        """Update the options of a plugin
        :param plugin_name: str of plugin name
        :param options: dict of options to be updated
        """
        xml_parent: ET.Element
        if isinstance(xml_parent_name, str):
            xml_parent = self._get_element_from_name(xml_parent_name)
        else:
            xml_parent = xml_parent_name
        plugin_elem: ET.Element = xml_parent.find("./plugin")
        self._add_plugin_options(plugin_elem, options)
        self._parse_XML()
        if save_changes:
            self.write_to_XML()

    def _add_plugin_options(self,
                            plugin_elem: ET.Element,
                            options
                            ):
        for key in options.keys():
            if isinstance(options[key], list):
                new_option = ET.SubElement(plugin_elem, key)
                option_text = ("\n{}".format(
                    "\n".join([*options[key]])))
                new_option.text = option_text
            elif isinstance(options[key], (int, float, str)):
                new_option = plugin_elem.find(str("./" + key))
                if new_option is None:
                    new_option = ET.SubElement(plugin_elem, key)
                text_lead = "\n" if "\n" not in str(options[key]) else ""
                new_option.text = "{0} {1}".format(
                    text_lead, str(options[key]))
            elif isinstance(options[key], (dict)):
                self._add_plugin_options(plugin_elem, options[key])

    def append_input_data(self,
                          data_name: str,
                          data_dir: str,
                          xml_parent: Union[ET.Element, str] = "Initialiser",
                          save_dir_as_relative: bool = True):
        """Appened path to input datafile. Replaces windows backslash
        :param plugin_type: string type of plugin to be loaded into module
        :param plugin_options: dict where keys & values are options & values
        :param xml_parent: dict OR str. 
                            If string given, parent elem is found via search,
                            Otherwise, plugin appeneded directly
        :param save_dir_as_relative: bool. If True [default], attempts to 
                            replace the library base path in the absolute
                            filename with "./" to make it relative to library
                            path. Recommended.
        """
        if isinstance(xml_parent, str):
            xml_parent = self._get_element_from_name(xml_parent)

        input_data_elem = xml_parent.find("./inputdata")

        if input_data_elem is None:
            input_data_elem = ET.SubElement(xml_parent, "inputdata")

        plugin_elem = ET.SubElement(input_data_elem, data_name)
        if save_dir_as_relative:
            data_dir = data_dir.replace(self.lib_base_path, "./")
        data_dir = data_dir.replace("\\", "/")
        plugin_elem.set('file', data_dir)

    def append_plugin_to_module(self,
                                plugin_type: str,
                                plugin_options: dict,
                                xml_parent: Union[ET.Element, str],
                                overwrite_existing: Union[bool, int] = False
                                ):
        """Appened plugin as subelement to existing module element
        :param plugin_type: string type of plugin to be loaded into module
        :param plugin_options: dict where keys & values are options & values
        :param xml_parent: dict OR str. 
                            If string given, parent elem is found via search,
                            Otherwise, plugin appeneded directly
        """
        if isinstance(xml_parent, str):
            xml_parent = self._get_element_from_name(xml_parent)

        plugin_elem = xml_parent.find("./plugin")

        if plugin_elem is not None and overwrite_existing:
            xml_parent.remove(plugin_elem)
            plugin_elem = None

        if plugin_elem is None:
            plugin_elem = ET.SubElement(xml_parent, "plugin")
            plugin_elem.set('type', plugin_type)
        self._add_plugin_options(plugin_elem, plugin_options)

    def append_pipeline_module(self,
                               module_type: str,
                               module_name: str,
                               plugin_type: str,
                               plugin_options: Dict[str, Any],
                               parents: list,
                               children: list,
                               xml_parent_element_name: str,
                               coords: list = None
                               ) -> None:
        """Append new pipeline module to existing XML elementTree to be written later
        :param module_type: string declare type of module (UserFeedback,data_processing etc)
        :param module_name: string give module a user-defined name
        :param plugin_type: string type of plugin to be loaded into module
        :param plugin_options: dict where keys & values are options & values
        :param parents: list of parent names for this module (can be empty)
        :param children: list of child names for this module (can be empty)
        :param xml_parent_element_name: str containing name of parent Element for new module
        :param coords [optional]: list of coordinates for UserFeedback canvas
        """
        xml_parent_element: ET.Element = self._get_element_from_name(
            xml_parent_element_name)

        new_mod = ET.Element(module_type.replace(" ", ""))
        new_mod.set('name', module_name)

        if plugin_type != None:
            self.append_plugin_to_module(plugin_type,
                                         plugin_options,
                                         new_mod,
                                         0
                                         )

        if xml_parent_element.tag == "loop":
            parents.append(xml_parent_element.attrib["name"])
        new_mod = self._add_relationships(new_mod, parents, children)
        new_mod = self._add_coords(new_mod, coords)

        self._loop_rels_autofill(xml_parent_element, module_name)
        xml_parent_element.append(new_mod)

    def append_pipeline_loop(self,
                             loop_type: str,
                             condition: str,
                             loop_name: str,
                             parents: list,
                             children: list,
                             xml_parent_element_name: str = None,
                             coords: list = None
                             ) -> None:
        """Append new pipeline module to existing XML file
        :param loop_type: string declare type of loop (for/while/manual etc)
        :param loop_name: string give loop a user-defined name
        :param plugin_type: string type of plugin to be loaded into module
        :param parents: list of parent names for this module (can be empty)
        :param children: list of child names for this module (can be empty)
        :param xml_parent_element_name: str containing name of parent Element for new module
        :param coords [optional]: list of coordinates for UserFeedback canvas
        """
        xml_parent_element: ET.Element = self._get_element_from_name(
            xml_parent_element_name)

        new_loop = ET.Element("loop")
        new_loop.set('type', loop_type)
        new_loop.set('condition', condition)
        new_loop.set('name', loop_name)

        if xml_parent_element.tag == "loop":
            parents.append(xml_parent_element.attrib["name"])
        new_loop = self._add_relationships(new_loop, parents, children)
        new_loop = self._add_coords(new_loop, coords)

        self._loop_rels_autofill(xml_parent_element, loop_name)

        xml_parent_element.append(new_loop)

    def _get_init_data_structure(self) -> Dict[str, Any]:
        data_struct = self._find_dict_with_key_val_pair(
            self.loaded_modules,
            "class", "data")

        assert len(data_struct) < 2, \
            "Multiple data with same ID, please check XML"

        out: Dict[str, Any]
        if len(data_struct) == 1:
            out = data_struct[0]
        else:
            out = {"to_load": {}}
        return out

    @property
    def data_to_load(self) -> Dict[str, str]:
        return (self._get_init_data_structure()["to_load"])


# Use case examples:
if __name__ == "__main__":
    s = XML_handler()
    s.load_XML("./examples/xml_files/regerssion_test.xml")
    s.update_plugin_options("Modelling-1",{"alpha":4})
    s._print_xml_config()
    a = s.data_to_load
    print(1)

    # s.write_to_XML()
    # s.append_pipeline_loop("for",
    #                       "10",
    #                       "my_loop_3",
    #                       ["Init"],
    #                       [])
    # s.append_pipeline_module("UserFeedback thing",
    #                       "added_mod",
    #                       "startpage",
    #                       {"class_list":["test_1","test_2"],"class_list_2":["test_1","test_2"]},
    #                       ["loop0"],
    #                       ["Output","loop0"],
    #                       "loop0",
    #                       [2,3,4,5])
    # s.write_to_XML()


from os import path

def import_plugin(script_config, plugin_name):
    plugin_list = __import__(script_config["__package__"]+'.plugins.' + plugin_name,
                             script_config,
                             {},
                             [plugin_name])
    plugin_class = getattr(plugin_list, plugin_name)
    return plugin_class

def import_plugin_absolute(script_config,plugin_package,plugin_name):
    plugin_list = __import__(plugin_package,
                             script_config,
                             {},
                             [plugin_name])
    plugin_class = getattr(plugin_list, plugin_name)
    return plugin_class

def import_module(script_config, module_name):
    module_list = __import__('vai_lab.' + module_name + "." + module_name + "_core",
                             script_config,
                             {},
                             [module_name])
    module_class = getattr(module_list, module_name)
    return module_class

def get_lib_parent_dir():
    """Returns the absolute path of the current library
    
    :returns: str of absolute path of the library root dir
    """
    return [__file__[:i] \
                for i,_ in enumerate(__file__)\
                if __file__[:i].\
                endswith("{0}vai_lab{0}".format(path.sep))][-1]

def rel_to_abs(filename: str) -> str:
        """Checks if path is relative or absolute
        If absolute, returns original path 
        If relative, converts path to absolute by appending to base directory
        """
        if filename[0] == ".":
            filename = path.join(get_lib_parent_dir(), filename)
        elif filename[0] == "/" or (filename[0].isalpha() and filename[0].isupper()):
            filename = filename
        return filename


from vai_lab._plugin_templates import DataProcessingT
from vai_lab._types import DataInterface

from sklearn.preprocessing import Normalizer as model
import pandas as pd

_PLUGIN_READABLE_NAMES = {"Normalizer": "default",
                          "Norm": "alias", "normalizer": "alias"}   # type:ignore
_PLUGIN_MODULE_OPTIONS = {"Type": "scaler"}                         # type:ignore
_PLUGIN_REQUIRED_SETTINGS = {"Data": "str"}                         # type:ignore
_PLUGIN_OPTIONAL_SETTINGS = {"norm": "str"}                         # type:ignore
_PLUGIN_REQUIRED_DATA = {}                                          # type:ignore
_PLUGIN_OPTIONAL_DATA = {"X", "Y", "X_tst", 'Y_tst'}                # type:ignore


class Normalizer(DataProcessingT):
    """
    Normalize samples individually to unit norm
    """

    def __init__(self):
        """Initialises parent class. 
            Passes `globals` dict of all current variables
        """
        super().__init__(globals())
        self.proc = model()

    def configure(self, config: dict):
        """Sets and parses plugin configurations options
        :param config: dict of internal tags set in the XML config file 
        """
        super().configure(config)

    def set_data_in(self, data_in):
        """Sets and parses incoming data
        :param data_in: saves data as class variable
                        expected type: vai_lab.Data.Data_core.Data
        """
        super().set_data_in(data_in)

    def fit(self):
        cleaned_options = self._clean_solver_options()
        self.proc.set_params(**cleaned_options)
        self.proc.fit(self.X)

    def transform(self, data: DataInterface) -> DataInterface:
        data.append_data_column("X", pd.DataFrame(self.proc.transform(self.X)))
        if self.X_tst is not None:
            data.append_data_column("X_test", pd.DataFrame(
                self.proc.transform(self.X_tst)))
        return data



from typing import Dict
from vai_lab._types import DataInterface
from abc import ABC, abstractmethod

import numpy as np


class PluginTemplate:
    def __init__(self, plugin_globals: dict) -> None:
        """Instantiates data containers and parses plugin-specific options

        param: plugin_globals:dict dictionary representing the global symbol table of the plugin script
        """
        self.X = None
        self.Y = None
        self.X_tst = None
        self.Y_tst = None

        self._PLUGIN_READABLE_NAMES: dict
        self._PLUGIN_MODULE_OPTIONS: dict
        self._PLUGIN_REQUIRED_SETTINGS: dict
        self._PLUGIN_OPTIONAL_SETTINGS: dict
        self._PLUGIN_REQUIRED_DATA: dict
        self._PLUGIN_OPTIONAL_DATA: dict

        self._parse_plugin_requirements(plugin_globals)

    def configure(self, config: dict) -> None:
        """Sets and parses plugin configurations options
        :param config: dict of internal tags set in the XML config file 
        """
        self._config = config
        self._parse_config()

    def set_data_in(self, data_in: DataInterface) -> None:
        """Sets and parses incoming data
        :param data_in: saves data as class variable
                        expected type: vai_lab.Data.Data_core.Data
        """
        req_check = [
            r for r in self._PLUGIN_REQUIRED_DATA if r not in data_in.keys()]
        if len(req_check) > 0:
            self._data_missing_error(req_check)
        self._data_in = data_in

    def _parse_plugin_requirements(self, plugin_globals: dict):
        """Get all plugin options from child plugin.
        """
        for key in plugin_globals.keys():
            if key.startswith("_PLUG"):
                self.__setattr__(key, plugin_globals[key])
        self._get_plugin_deafult_name(plugin_globals)

    def _get_plugin_deafult_name(self, plugin_globals: dict):
        """Parse self._PLUGIN_READABLE_NAMES for default name and set to self.default_name"""
        self.default_name = None
        for name in self._PLUGIN_READABLE_NAMES:
            if self._PLUGIN_READABLE_NAMES[name] == "default":
                if self.default_name is not None:
                    print("Multiple deafult names specified in " +
                          plugin_globals["__file__"])
                self.default_name = name

    def _data_missing_error(self, req):
        """If incoming data does not match required data, raise Exception and print"""
        raise Exception("Minimal Data Requirements not met"
                        + "\n\t{0} ".format(self.default_name)
                        + "requires data: {0}".format(self._PLUGIN_REQUIRED_DATA)
                        + "\n\tThe following data is missing:"
                        + "\n\t\u2022 {}".format(",\n\t\u2022 ".join([*req])))

    def _parse_config(self):
        """Parse incoming data and args, sets them as class variables"""
        self.X = np.array(self._get_data_if_exist(self._data_in, "X"))
        self.Y = np.array(self._get_data_if_exist(self._data_in, "Y")).ravel()
        self.X_tst = self._get_data_if_exist(self._data_in, "X_test")
        self.Y_tst = np.array(self._get_data_if_exist(
            self._data_in, "Y_test")).ravel()
        self._clean_options()

    def _get_data_if_exist(self, data_dict: dict, key: str, default=None):
        """Returns data from incoming data if exists, else returns None

        :param data_dict:dict of all incoming data
        :param key:str: name of data being searched for

        :returns result:dict OR None: dict of data if exists, else None
        """
        return data_dict[key] if key in data_dict.keys() and data_dict[key] is not None else default

    def _reshape(self, data, shape):
        """Reshapes data to shape. Primarily used for arrays with shape (n,)

        :param data: data to reshape
        :param shape: shape to reshape to

        :returns reshaped data: reshaped data
        """
        return data.reshape(shape[0], shape[1])

    def _parse_options_dict(self,options_dict:Dict):
        for key, val in options_dict.items():
            if type(val) == str and val.replace('.', '').replace(',', '').isnumeric():
                cleaned_opts = []
                for el in val.split(","):
                    val = float(el)
                    if val.is_integer():
                        val = int(val)
                    cleaned_opts.append(val)
                options_dict[key] = cleaned_opts
        return options_dict

    def _clean_options(self):
        """Parses incoming plugin options in self._config["options"] 
                and modifies DataInterface in-place
                str options which only contain numeric data are converted to float OR int
        """
        self._parse_options_dict(self._config["options"])

    def _test(self, data: DataInterface) -> DataInterface:
        """Run debug tests on data operations
        TODO: Investigate if all plugins need a score and predict method
        """
        if self._PLUGIN_MODULE_OPTIONS['Type'] == 'classification':
            print('Training accuracy: %.2f%%' %
                  (self.score(self.X, self.Y)*100))  # type: ignore
            if self.Y_tst is not None:
                print('Test accuracy: %.2f%%' %
                      (self.score(self.X_tst, self.Y_tst)*100))
            if self.X_tst is not None:
                data.append_data_column("Y_pred", self.predict(self.X_tst))
            return data
        elif self._PLUGIN_MODULE_OPTIONS['Type'] == 'regression':
            print('Training R2 score: %.3f' %
                  (self.score(self.X, self.Y)))  # type: ignore
            if self.Y_tst is not None:
                print('Training R2 score: %.3f' %
                      (self.score(self.X_tst, self.Y_tst)))
            if self.X_tst is not None:
                data.append_data_column("Y_pred", self.predict(self.X_tst))
            return data
        elif self._PLUGIN_MODULE_OPTIONS['Type'] == 'clustering':
            print('Clustering completed')
            if self.X_tst is not None:
                data.append_data_column("Y_pred", self.predict(self.X_tst))
            return data
        else:
            return data


class DataProcessingT(PluginTemplate, ABC):
    def __init__(self, plugin_globals: dict) -> None:
        super().__init__(plugin_globals)
        self._options_to_ignore = ["Data"]

    def configure(self, config: dict):
        """Extended from PluginTemplate.configure"""
        super().configure(config)
        if type(self.X) is None and type(self.Y) is None:
            print('Invalid Data name. Indicate whether to use `X` or `Y`')

    def _clean_solver_options(self):
        """XML tags that specify data to be processed should not be sent to solver
        TODO: in future, reverse the direction of this function - instead pull all options
                that are directly required by the solver
        """
        return {i[0]: i[1]
                for i in self._config["options"].items()
                if i[0] not in self._options_to_ignore}

    @abstractmethod
    def fit(self):
        pass

    @abstractmethod
    def transform(self, data: DataInterface) -> DataInterface:
        pass


class ModellingPluginT(PluginTemplate, ABC):
    def __init__(self, plugin_globals: dict) -> None:
        super().__init__(plugin_globals)

    @abstractmethod
    def solve(self):
        pass

    @abstractmethod
    def predict(self):
        pass

    @abstractmethod
    def score(self):
        pass


class UI(PluginTemplate, ABC):
    @property
    def class_list(self):
        pass
        # """Set the class list"""

    @class_list.setter  # type: ignore
    @abstractmethod
    def class_list(self, value):
        pass

    @abstractmethod
    def save_file(self):
        pass

    @abstractmethod
    def save_file_as(self):
        pass


class EnvironmentPluginT(PluginTemplate, ABC):
    
    @abstractmethod
    def load_model(self) -> None:
        pass

    @abstractmethod
    def connect(self):
        pass

    @abstractmethod
    def reset(self):
        pass

    @abstractmethod
    def disconnect(self):
        pass

    @abstractmethod
    def set_gui(self,use_gui:bool=True) -> None:
        pass

    def configure(self, config: dict):
        """Extended from PluginTemplate.configure"""
        super().configure(config)
        self.set_gui(self._config["options"]["headless"])


from vai_lab._plugin_templates import ModellingPluginT
from sklearn.neighbors import KNeighborsClassifier as model

_PLUGIN_READABLE_NAMES = {"KNNClassifier": "default",
                          "KNN-C": "alias"}             # type:ignore
_PLUGIN_MODULE_OPTIONS = {"Type": "classification"}     # type:ignore
_PLUGIN_REQUIRED_SETTINGS = {}                          # type:ignore
_PLUGIN_OPTIONAL_SETTINGS = {"n_neighbors": "int",
                             "weights": "str"}          # type:ignore
_PLUGIN_REQUIRED_DATA = {"X", "Y"}                      # type:ignore
_PLUGIN_OPTIONAL_DATA = {"X_tst", 'Y_tst'}              # type:ignore


class KNNclassifier(ModellingPluginT):
    """
    Classifier implementing the k-nearest neighbors vote
    """

    def __init__(self):
        """Initialises parent class. 
            Passes `globals` dict of all current variables
        """
        super().__init__(globals())
        self.clf = model()

    def configure(self, config: dict):
        """Sets and parses plugin configurations options
        :param config: dict of internal tags set in the XML config file 
        """
        super().configure(config)

    def set_data_in(self, data_in):
        """Sets and parses incoming data
        :param data_in: saves data as class variable
                        expected type: vai_lab.Data.Data_core.Data
        """
        super().set_data_in(data_in)

    def solve(self):
        """Sends params to solver, then runs solver"""
        self.clf.set_params(**self._config["options"])
        self.clf.fit(self.X, self.Y)

    def predict(self, data):
        """Uses fitted model to predict output of a given Y
        :param data: array-like or sparse matrix, shape (n_samples, n_features)
                    Samples
                    expected type: vai_lab.Data.Data_core.Data
        :returns: array, shape (n_samples,)
                    Returns predicted values.
        """
        return self.clf.predict(data)

    def score(self, X, Y, sample_weight=None):
        """Return the coefficient of determination
        :param  X : array-like of shape (n_samples, n_features)
        :param  Y :  array-like of shape (n_samples,) or (n_samples, n_outputs)
        :param sample_weight : array-like of shape (n_samples,), default=None
                    Sample weights.

        :returns: score : float R^2` of ``self.predict(X)`` wrt. `y`.
        """
        return self.clf.score(X, Y, sample_weight=None)


from vai_lab.Data.xml_handler import XML_handler
from vai_lab._import_helper import import_plugin_absolute
from vai_lab._types import DataInterface, PluginSpecsInterface, DictT

from typing import Any
import tkinter as tk
from tkinter.font import Font


class GUI(tk.Tk):
    """
    TODO: This structure still needs serious overhaul. 

    TODO: By having the TKinter controller as in the main GUI module, we are locked to using TKinter which defeats the purpose of being modular. This class needs to only be calling the plugins themselves, not acting as a controller.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.title_font = Font(family='Helvetica',
                                      size=14,
                                      weight="bold")
        self.pages_font = Font(family='Helvetica',
                                      size=12)
        self._desired_ui_types = []
        self._top_ui_layer = None
        self._module_config = None
        self._debug = False
        self.closed = False
        self.startpage = False
        self.output = {}

    def set_avail_plugins(self, avail_plugins: PluginSpecsInterface):
        self._avail_plugins = avail_plugins

    def set_data_in(self, data_in: DataInterface):
        self._data_in = data_in

    def set_options(self, module_config: DictT):
        """Send configuration arguments to GUI

        :param module_config: dict of settings to congfigure the plugin
        """
        self._module_config = module_config
        self._load_plugin(self._module_config["plugin"]["plugin_name"])

    def set_gui_as_startpage(self):
        self.startpage = True
        self._load_plugin("main")
        self.s = XML_handler()
        self.s.new_config_file()
        
    def set_gui(self, status):
        self.startpage = True
        self._load_plugin("progressTracker")
        self.status = status

    def _compare_layer_priority(self, ui_specs):
        """Check if a new module should have higher layer priority than the existing one

        :param ui_name: name of the UI method being compared
        :type ui_name: str
        """
        if self._top_ui_layer == None:
            self._top_ui_layer = ui_specs
        else:
            current_top_layer = self._top_ui_layer["_PLUGIN_MODULE_OPTIONS"]["layer_priority"]
            candidate_layer = ui_specs["_PLUGIN_MODULE_OPTIONS"]["layer_priority"]
            self._top_ui_layer = candidate_layer \
                if candidate_layer < current_top_layer \
                else self._top_ui_layer

    def _add_UI_type_to_frames(self, ui_specs):
        """Add user defined UI method to list of frames to be loaded

        :param ui_name: name of the UI method being loaded
        :type ui_name: str 
        """
        plugin = import_plugin_absolute(globals(),
                                        ui_specs["_PLUGIN_PACKAGE"],
                                        ui_specs["_PLUGIN_CLASS_NAME"])

        self._desired_ui_types.append(plugin)
        self._compare_layer_priority(ui_specs)
        if ui_specs["_PLUGIN_MODULE_OPTIONS"]["required_children"] != None:
            for children in ui_specs["_PLUGIN_MODULE_OPTIONS"]["required_children"]:
                self._load_plugin(children)

    def _load_plugin(self, ui_type: list):
        """"Given user input, create a list of classes of the corresponding User Interface Type 

        :param ui_name: name of the desired User Interface Method
        :type ui_name: str or list
        """
        ui_type = ui_type\
            if isinstance(ui_type, list)\
            else [ui_type]

        for ui in ui_type:
            ui_specs = self._avail_plugins.find_from_readable_name(ui)
            try:
                self._add_UI_type_to_frames(ui_specs)
            except ModuleNotFoundError as ex:
                from sys import exit
                print(ex.msg)
                exit(1)
            except:
                from sys import exit
                print(
                    "Error: User Interface \"{0}\" not recognised. \
                    \nAvailable methods are: \
                    \n  - {1}"
                    .format(ui, ",\n  - ".join(
                        [i for i in self._avail_plugins.available_plugin_names])))
                exit(1)

    def _append_to_output(self, key: str, value: Any):
        self.output[key] = value

    def _show_frame(self, page_name):
        '''Show a frame for the given page name'''
        if isinstance(page_name, dict):
            page_name = page_name["_PLUGIN_CLASS_NAME"]
        frame = self.frames[page_name]
        frame.tkraise()

    def _on_closing(self) -> None:
        self.closed = True
        self.destroy()

    def destroy(self) -> None:
        return super().destroy()

    def launch(self):
        """Runs UserInterface Plugin. 
        If multiple frames exist, they are stacked
        """
        container = tk.Frame(self, bg='#064663')
        container.pack(side="top", fill="both", expand=True)
        container.grid_rowconfigure(0, weight=1)
        container.grid_columnconfigure(0, weight=1)
        self.output_data= []
        self.frames = {}
        for F in self._desired_ui_types:
            page_name = F.__name__
            frame = F(parent=container, controller=self,
                      config=self._module_config)
            if not self.startpage:
                frame.set_data_in(self._data_in)
            self.frames[page_name] = frame

            frame.grid(row=0, column=0, sticky="nsew")
            self.output_data.append(frame.out_data)

        self._show_frame(self._top_ui_layer)
        self.protocol("WM_DELETE_WINDOW", self._on_closing)
        if not self._debug:
            self.mainloop()
        else:
            self.closed = True
        return self.output

    def get_result(self):
        """TODO: This currently does not work, as the output data is stored in a list
        This is due to the way this module is constructed, it cycles through plugins, 
        which is useful for the startpage, but complicates things when running
        UserInterface plugins"""
        return self.output_data